from .expand import expand
from .local_rst import local_rst
from .factor import factor
from .flatten import flatten
from .project import project
from .czip import czip, zip_sync, unzip
from .ccat import ccat
from .cart import cart, cart_sync, cart_sync_with
from .union import union_sync, union_collapse, union_sync_with, pick_by, filt_by, do_if
from .queue import queue_wrap_from
from .cast import cast
from .quenvelope import quenvelope
from .sieve import sieve
from .fmap import fmap
from .permute import permuted_apply
from .const import const
from .add import add
from .sub import sub
from .mul import mul
from .div import div
from .mod import mod
from .neg import neg
from .eq import eq
from .neq import neq
from .lt import lt
from .invert import invert
from .mux import mux, mux_zip, mux_valve
from .demux import demux, demux_zip, demux_by
from .shred import shred
from .decoupler import decoupler, buff
from .fifo import fifo
from .fill import fill
from .dreg import dreg
from .serialize import serialize
from .data_dly import data_dly
from .filt import filt
from .shr import shr
from .align import align

import pygears.common.fmaps

__all__ = [
    'expand', 'czip', 'ccat', 'cart', 'cast', 'quenvelope', 'zip_sync',
    'sieve', 'flatten', 'fmap', 'permuted_apply', 'const', 'add', 'sub', 'mul',
    'div', 'neg', 'mux', 'demux', 'shred', 'cart_sync', 'decoupler', 'dreg',
    'unzip', 'serialize', 'project', 'fifo', 'factor', 'fill', 'mux_zip',
    'mux_valve', 'demux_zip', 'data_dly', 'eq', 'union_sync', 'union_collapse',
    'filt', 'buff', 'local_rst', 'mod', 'invert', 'shr', 'queue_wrap_from',
    'demux_by', 'align', 'lt', 'neq'
]



from pygears.registry import PluginBase
from pygears import registry, find
from pygears.rtl.inst import rtl_inst
from pygears.rtl.connect import rtl_connect
from pygears.rtl.channel import RTLChannelVisitor, RTLOutChannelVisitor


def rtlgen(top=None, **conf):

    if top is None:
        top = registry('HierRoot')
    elif isinstance(top, str):
        top = find(top)

    for oper in registry('RTLFlow'):
        top = oper(top, conf)

    return top


class RTLPlugin(PluginBase):
    @classmethod
    def bind(cls):
        cls.registry['RTLFlow'] = [
            rtl_inst, rtl_connect, RTLChannelVisitor, RTLOutChannelVisitor
        ]



from pygears.core.gear import gear
from pygears.core.intf import IntfOperPlugin
from pygears import module


@gear
async def sieve(din, *, index) -> b'din[index]':
    async with din as d:
        dout = []
        for i in index:
            dout.append(d[i])

        if len(index) == 1:
            dout = dout[0]

        yield module().tout(dout)


def getitem(self, index):
    index = self.dtype.index_norm(index)
    name_appendices = []
    for i in index:
        if isinstance(i, slice):
            name_appendices.append(f'{i.start}v{i.stop}')
        else:
            name_appendices.append(f'{i}')

    return self | sieve(index=index, name='sieve_' + '_'.join(name_appendices))


class GetitemIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace']['__getitem__'] = getitem



#!/usr/bin/env python3

import contextlib
import select
import socket
import pexpect
import logging
import argparse
import os
import re
import sys


def set_keepalive_linux(sock, after_idle_sec=1, interval_sec=1, max_fails=3):
    """Set TCP keepalive on an open socket.

    It activates after 1 second (after_idle_sec) of idleness,
    then sends a keepalive ping once every 3 seconds (interval_sec),
    and closes the connection after 5 failed ping (max_fails), or 15 seconds
    """
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, after_idle_sec)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, interval_sec)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, max_fails)


def client_loop(client, p):
    while 1:
        data_sent = None
        ready = select.select([client], [], [], 0.01)

        if ready[0]:
            try:
                data_sent = client.recv(4096)
                if data_sent:
                    data_sent = data_sent.decode()
                    logging.debug(f'Received client data: {data_sent}')
                    p.send(data_sent)
                else:
                    logging.info(
                        f'EOF received at client socket {client.fileno()}')
                    return
            except ConnectionResetError:
                return

        data = ''
        iters = 0
        try:
            while (1):
                data += p.read_nonblocking(
                    size=4096, timeout=0.01 if (data_sent is None) else 0.1)
                iters += 1

        except pexpect.TIMEOUT:
            if data:
                lines = data.split('\r\n')
                for i in range(len(lines)):
                    if lines[0].endswith("% "):
                        break
                else:
                    i = 0
                data = '\n'.join(lines[i:])
                if data_sent:
                    data = data.replace(data_sent.replace('\r', '\n'), '')
                logging.debug(f'Sending to client: {repr(data)}')
                client.sendall(data.encode())
        except Exception as e:
            logging.exception("Pexpect exception")
            raise


def open_socket(port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(('', port))
    sock.listen(1)
    set_keepalive_linux(sock)

    return sock


def server_loop(port, log_level='INFO'):
    logging.basicConfig(
        filename=os.path.expanduser(f'~/pywave_{port}.log'),
        level=getattr(logging, log_level),
        format='%(asctime)s %(levelname)-8s %(message)s')

    logging.info(f'Pywave process started, pid={os.getpid()}')

    try:
        sock = open_socket(port)
    except socket.error as err:
        logging.exception("Socket error")
        raise

    logging.info(f'Server socket {sock.fileno()} opened at port {port}.')

    try:

        p = pexpect.spawnu('gtkwave -W')
        p.expect('%')
        version = re.search(r"GTKWave Analyzer v(\d{1}\.\d{1}.\d{2})",
                            p.before).group(0)

        logging.info(f'GTKWave {version} started, listening on port {port}...')

        while 1:
            client, address = sock.accept()
            logging.info(f'Client connected at socket {client.fileno()}')
            client.setblocking(0)

            try:
                client_loop(client, p)
            except pexpect.EOF:
                logging.info('GTKWave terminated unexpectedly. Exiting...')
                return
            except Exception:
                raise
            finally:
                logging.info(
                    f'Client at socket {client.fileno()} disconnected.')
                client.close()

    except Exception:
        sock.close()
        # p.terminate()
        raise


def main(argv=sys.argv):
    parser = argparse.ArgumentParser(prog="GTKWave deamon TCP server")

    parser.add_argument(
        '-p', dest='port', type=int, default=60000, help="Server port")

    parser.add_argument(
        '-l', dest='log_level', default='INFO', help="Logging level")

    args = parser.parse_args()

    with open(os.path.expanduser(f'~/pywave_{args.port}.stdout'), 'w') as f:
        with contextlib.redirect_stdout(f), contextlib.redirect_stderr(f):
            server_loop(args.port, args.log_level)


if __name__ == '__main__':
    main()



import os
import sys
from enum import IntEnum
from functools import partial
from traceback import (extract_stack, extract_tb, format_exception_only,
                       format_list, walk_stack, walk_tb)

from pygears.registry import PluginBase, registry


class ErrReportLevel(IntEnum):
    debug = 0
    user = 1


class ErrReportPlugin(PluginBase):
    @classmethod
    def bind(cls):
        # cls.registry['ErrReportLevel'] = ErrReportLevel.user
        # from .pdb_patch import patch_pdb
        # patch_pdb()

        cls.registry['ErrReportLevel'] = ErrReportLevel.debug
        cls.registry['ExitHooks'] = []


def register_exit_hook(hook, *args, **kwds):
    registry('ExitHooks').append(partial(hook, *args, **kwds))


def parse_trace(s, t):
    if registry("ErrReportLevel") == ErrReportLevel.debug:
        yield s
    else:
        is_internal = t[0].f_code.co_filename.startswith(
            os.path.dirname(__file__))
        is_boltons = 'boltons' in t[0].f_code.co_filename
        if not is_internal and not is_boltons:
            yield s


def enum_traceback(tr):
    for s, t in zip(format_list(extract_tb(tr)), walk_tb(tr)):
        yield from parse_trace(s, t)


def enum_stacktrace():
    for s, t in zip(format_list(extract_stack()), walk_stack(f=None)):
        yield from parse_trace(s, t)


def pygears_excepthook(exception_type,
                       exception,
                       tr,
                       debug_hook=sys.excepthook):

    for hook in registry('ExitHooks'):
        try:
            hook()
        except:
            pass

    if registry("ErrReportLevel") == ErrReportLevel.debug:
        debug_hook(exception_type, exception, tr)
    else:
        from pygears.util.print_hier import print_hier
        from pygears import find

        try:
            print_hier(find('/'))
        except Exception as e:
            pass

        # print traceback for LogException only if appropriate
        # 'print_traceback' in registry is set
        from pygears.core.log import LogException, registry_log_name
        print_traceback = (exception_type is not LogException)
        if not print_traceback:
            print_traceback = registry(
                registry_log_name.substitute(
                    name=exception.name))['print_traceback']
        if print_traceback:
            for s in enum_traceback(tr):
                print(s, end='')

        print(format_exception_only(exception_type, exception)[0])



from pygears import gear
from pygears.typing import Queue, Tuple, Uint


@gear
async def replicate(din: Tuple[Uint['w_len'], 'w_val']) -> Queue['w_val']:
    async with din as val:
        for i in range(val[0]):
            yield (val[1], i == (val[0] - 1))



from pygears.core.gear import gear
from pygears.typing import Queue, Unit
from pygears import module
from pygears.util.utils import quiter_async


@gear(sv_param_kwds=[], enablement=b'din_lvl >= lvl')
async def quenvelope(din: Queue['din_t', 'din_lvl'], *,
                     lvl) -> Queue[Unit, 'lvl']:
    """Extracts the queue structure of desired level called the envelope

    If there are more eot levels then forwarded to the output, those eot excess
levels are called subenvelope (which is not passed to the output). When
there is a subenvelope, the number of data the output transactions (envelope)
will contain is lowered by contracting each input transactions within
subenvelope to the length of 1. This is done in order that the envelope can be
correctly used within cartesian concatenations.

    """

    dout = module().dout
    sub_lvl = din.dtype.lvl - lvl
    out_data = None

    async for data in quiter_async(din):
        if out_data is None:
            out_data = module().tout((Unit(), *data[-lvl:]))
            dout.put_nb(out_data)

        if sub_lvl > 0:
            subelem = data[:len(data)-lvl]
            if subelem.last:
                out_data = None
        else:
            out_data = None

        await dout.ready()



from pygears import gear, alternative
from pygears.typing import Queue, Union


def filt_type(din, lvl, sel):
    return Queue[(din[0].types)[sel], lvl]


@gear
def filt(din: Union, *, sel) -> b'din.types[sel]':
    pass


@alternative(filt)
@gear(svgen={'svmod_fn': 'qfilt.sv'})
def qfilt(din: Queue['TUnion', 'lvl'],
          *,
          sel=0,
          filt_lvl=1,
          w_din=b'int(din[0][0])',
          w_dout=b'int((din[0].types)[sel])') -> b'filt_type(din, lvl, sel)':
    pass



from pygears.svgen.svmod import SVModuleGen
from pygears.typing.queue import Queue
from pygears.svgen.inst import SVGenInstPlugin
from pygears.common.demux import demux_zip
from .syncguard import SVGenSyncGuard


class SVGenDemuxZip(SVModuleGen):
    def __init__(self, node):
        super().__init__(node)
        self.syncguard = SVGenSyncGuard(f'{self.sv_module_name}_syncguard', 2)

    @property
    def is_generated(self):
        return True

    @property
    def sv_file_name(self):
        return super().sv_file_name, self.syncguard.sv_file_name

    def get_module(self, template_env):
        context = {
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs())
        }
        return template_env.render_local(
            __file__, "demux_zip.j2", context), \
            self.syncguard.get_module(template_env)


class SVGenDemuxZipPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][demux_zip] = SVGenDemuxZip



from pygears.svgen.inst import SVGenInstPlugin
from pygears.svgen.svmod import SVModuleGen
from pygears.common.align import align


class SVGenAlign(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        params = {}

        from pygears.typing import Queue
        for l, p in zip(self.node.params['lvl'], self.node.in_ports):
            dtype = p.dtype
            if issubclass(dtype, Queue):
                dtype = dtype[0]
            params[f'W_{p.basename.upper()}_LVL'] = l
            params[f'W_{p.basename.upper()}_DATA'] = int(dtype)

        context = {
            'module_name': self.sv_module_name,
            'params': params,
            'intfs': list(self.sv_port_configs())
        }
        return template_env.render_local(__file__, "align.j2", context)


class SVGenAlignPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][align] = SVGenAlign



class SimEvent(list):
    """Simulator Event that can trigger list of callbacks.

    Event is implemented as a list of callable objects - callbacks.
    Calling an instance of this will cause a call to each item in
    the list in ascending order by index.

    Callback function should return a boolean value. If it returns:

    True    -- Callback is re-registered by the _event
    False   -- Callback is deleted from the list

    Callback can be registered with or without arguments. Callback
    without arguments is registered by adding function reference
    to the list. Callback with arguments is registered by adding
    a tuple to the list. The first tuple item contains function
    reference. The rest of the items will be passed to the
    callback once the _event is triggered.
    """

    def __call__(self, *args, **kwargs):
        """Trigger the _event and call the callbacks.

        The arguments passed to this function will be passed to
        all the callbacks.
        """

        expired = []

        for i, f in enumerate(self):
            # If additional callback arguments are passed
            if isinstance(f, tuple):
                func = f[0]
                fargs = f[1:]

                ret = func(*(fargs + args), **kwargs)
            else:
                ret = f(*args, **kwargs)

            # If callback should not be re-registered
            if not ret:
                expired.append(i)

        # Delete from the list all callback that returned false
        for e in reversed(expired):
            del self[e]

    def __repr__(self):
        return "Event(%s)" % list.__repr__(self)



from .base import EnumerableGenericMeta
from .unit import Unit


class ArrayMeta(EnumerableGenericMeta):
    def keys(self):
        """Returns a list of keys that can be used for indexing :class:`Array` [T, N] type. Number of keys equals to the number of elements N.

        >>> Array[Uint[2], 5].keys()
        [0, 1, 2, 3, 4]
        """

        return list(range(int(self.args[1])))

    # def __new__(cls, name, bases, namespace, args=[]):
    #     return super().__new__(cls, name, bases, namespace, args)

        # if len(cls.args) < 2:
        #     # Generic parameter values have not been supplied
        #     return cls
        # else:
        #     # If Array of Units, return Unit
        #     if cls.args[0] == Unit:
        #         return Unit

        #     if cls.args[1] == 1:
        #         # If Array of length 1, return original type
        #         return cls.args[0]
        #     elif cls.args[1] == 0:
        #         # If Array of length 0, return Unit
        #         return Unit

        #     # Otherwise, return the Array class
        #     return cls

    @property
    def dtype(self):
        return self.args[0]

    def __getitem__(self, index):
        """If a single element is supplied for index, returns type T. If a slice is suplied for index, an :class:`Array` type is returned with a number of elements equal to the slice size.

        >>> Array[Uint[2], 5][3]
        Uint[2]

        >>> Array[Uint[2], 5][2:4]
        Array[Uint[2], 2]
        """

        if not self.is_specified():
            return super().__getitem__(index)

        index = self.index_norm(index)

        if len(index) == 1 and not isinstance(index[0], slice):
            if index[0] >= len(self):
                raise IndexError

            return self.args[0]
        else:
            width = 0
            for i in index:
                if isinstance(i, slice):
                    if (i.stop == 0) or (i.stop - i.start > len(self)):
                        raise IndexError
                    width += i.stop - i.start
                else:
                    if i >= len(self):
                        raise IndexError
                    width += 1

            return Array[self.args[0], width]

    def __str__(self):
        return f'Array[{str(self.args[0])}, {len(self)}]'


class Array(tuple, metaclass=ArrayMeta):
    """Generic container datatype that holds N instances of type T

    Generic parameters:
       T: Type of the :class:`Array` [T, N] elements

       N: Number of elements in the :class:`Array` [T, N]

    Concrete data type is obtained by indexing::

        u16_4 = Array[Uint[16], 4]

    """
    __parameters__ = ['T', 'N']

    def __new__(cls, val: tuple):
        array_tpl = (cls[0](v) for v in val)
        return super(Array, cls).__new__(cls, array_tpl)

    def __int__(self):
        w_dtype = int(type(self).dtype)
        ret = 0
        for d in reversed(self):
            ret <<= w_dtype
            ret |= int(d)

        return ret

    @classmethod
    def decode(cls, val):
        ret = []
        mask = int(len(cls[0]) * '1', 2)
        for t in cls:
            ret.append(t.decode(val & mask))
            val >>= int(t)

        return cls(tuple(ret))



"""Implements fixed width integer types: :class:`Uint` [N] - unsigned, :class:`Int` [N] - signed and :class:`Integer` [N] - sign agnostic type. These types correspond to HDL logic vector types.

Objects of these classes can also be instantiated and they provide some integer arithmetic capabilities.
"""

from pygears.typing.base import EnumerableGenericMeta, GenericMeta, typeof
from pygears.typing.tuple import Tuple
from pygears.typing.bool import Bool
from pygears.typing.bitw import bitw
from pygears.core.log import typing_log


class IntegerMeta(EnumerableGenericMeta):
    """Defines common methods for all Integer based classes.
    """

    def __str__(self):
        if self.args:
            if isinstance(self.args[0], int):
                return f'Z{self.args[0]}'
            else:
                return f'Z({self.args[0]})'
        else:
            return super().__str__()

    def __int__(self):
        return int(self.__args__[0])

    def __gt__(self, others):
        return int(self) > int(others)

    def keys(self):
        """Returns a list of keys that can be used for indexing the type.

        >>> Int[8].keys()
        [0, 1, 2, 3, 4, 5, 6, 7]
        """
        return list(range(int(self)))

    def __add__(self, other):
        """Returns the same type, but one bit wider to accomodate potential overflow.

        >>> Uint[8] + Uint[8]
        Uint[9]
        """
        return self.base[max(int(self), int(other)) + 1]

    __radd__ = __add__

    def __sub__(self, other):
        """Returns the signed Int type, but one bit wider to accomodate potential overflow.

        >>> Uint[8] + Uint[8]
        Int[9]
        """
        return Int[max(int(self), int(other)) + 1]

    def __mul__(self, other):
        """Returns the same type, whose width is equal to the sum of operand widths.

        >>> Uint[8] + Uint[8]
        Uint[16]
        """
        return self.base[int(self) + int(other)]

    def __truediv__(self, other):
        return self.base[int(self) - int(other) + 1]

    def __rtruediv__(self, other):
        return self.base[int(other) - int(self) + 1]

    def __floordiv__(self, other):
        return self.base[int(self) - int(other) + 1]

    def __rfloordiv__(self, other):
        return self.base[int(other) - int(self) + 1]

    def __mod__(self, other):
        return self.base[int(self) % int(other)]

    def __rmod__(self, other):
        return self.base[int(other) % int(self)]

    __rmul__ = __mul__

    def __getitem__(self, index):
        if not self.is_specified():
            return super().__getitem__(index)

        index = self.index_norm(index)

        width = 0
        for i in index:
            if isinstance(i, slice):
                if (i.stop == 0) or (i.stop - i.start > len(self)):
                    raise IndexError
                width += i.stop - i.start
            else:
                if i >= len(self):
                    raise IndexError
                width += 1

        return self.base[width]


def check_width(val, res):
    if not isinstance(val, tuple):
        val = (val, )

    total_width = 0
    for v in val:
        total_width += bitw(v)

    if (total_width > res.width):
        typing_log().warning(
            f'Value overflow - value {val} cannot be represented with {res.width} bits'
        )


class Integer(int, metaclass=IntegerMeta):
    """Base type for both :class:`Int` [N] and :class:`Uint` [N] generic types. Corresponds to HDL logic vector types. For an example Integer[9] translates to :sv:`logic [8:0]`.
    """

    def __new__(cls, val: int = 0):
        if type(val) == cls:
            return val

        res = super(Integer, cls).__new__(cls,
                                          int(val) & ((1 << len(cls)) - 1))
        check_width(val, res)
        return res

    @property
    def width(self):
        """Returns the number of bits used for the representation

        >>> Integer[8](0).width
        8
        """
        return len(type(self))

    def __len__(self):
        """Returns the number of bits used for the representation

        >>> len(Integer[8](0))
        8
        """
        return self.width

    def __add__(self, other):
        if isinstance(other, Integer):
            return (type(self) + type(other))(int(self) + int(other))
        else:
            return type(self)(int(self) + other)

    def __mul__(self, other):
        if isinstance(other, Integer):
            return (type(self) * type(other))(int(self) * int(other))
        else:
            return type(self)(int(self) * other)

    def __str__(self):
        return f'{str(type(self))}({int(self)})'

    def __repr__(self):
        return f'{repr(type(self))}({int(self)})'

    def __int__(self):
        """Returns builtin integer type

        >>> type(int(Integer[8](0)))
        <class 'int'>
        """
        return super(Integer, self).__int__()

    def __getitem__(self, index):
        """Returns the value of the indexth bit in the number representation.

        >>> Integer[8](0b10101010)[5]
        1
        """
        if isinstance(index, slice):
            bits = tuple(
                Bool(int(self) & (1 << i))
                for i in range(*index.indices(self.width)))

            return Uint[len(bits)](Tuple[(Bool, ) * len(bits)](bits))
        elif index < self.width:
            return Bool(int(self) & (1 << index))
        else:
            raise IndexError

    @classmethod
    def decode(cls, val):
        """Creates Integer object from any int-convertible object val.

        >>> Integer[8].decode(0xffff)
        Integer[8](255)
        """
        return cls(int(val))


class IntMeta(IntegerMeta):
    def __str__(self):
        if self.args:
            if isinstance(self.args[0], int):
                return f'i{self.args[0]}'
            else:
                return f'i({self.args[0]})'
        else:
            return super().__str__()


class Int(Integer, metaclass=IntMeta):
    """Fixed width generic signed integer data type.

    Generic parameters:
       N: Bit width of the :class:`Int` [N] representation

    Args:
       val: Integer value to convert to :class:`Int` [N]

    :class:`Int` [N] is a generic datatype derived from :class:`Integer` [N]. It represents signed integers with fixed width binary representation. Concrete data type is obtained by indexing:

    >>> i16 = Int[16]

    """
    __parameters__ = ['N']

    def __new__(cls, val: int = 0):
        if type(val) == cls:
            return val

        if cls.is_generic():
            if isinstance(val, Uint):
                return cls[val.width + 1](int(val))
            else:
                return cls[val.bit_length() + 1](int(val))
        else:
            return super(Int, cls).__new__(cls,
                                           int(val) & ((1 << len(cls)) - 1))

    def __int__(self):
        val = super(Int, self).__int__()
        if val >= (1 << (self.width - 1)):
            val -= 1 << self.width

        return val

    def __eq__(self, other):
        return int(self) == int(other)


class UintMeta(IntegerMeta):
    def __sub__(self, other):
        """Returns a Tuple of the result type and overflow bit.

        >>> Uint[16] - Uint[8]
        Tuple(Uint[16], Bool)
        """
        if (issubclass(other, Uint)):
            return Tuple[Uint[max(int(self), int(other))], Bool]
        else:
            return super().__sub__(self, other)

    def __str__(self):
        if not self.args:
            return f'u'
        elif isinstance(self.args[0], int):
            return f'u{self.args[0]}'
        else:
            return f'u({self.args[0]})'


class Uint(Integer, metaclass=UintMeta):
    """Fixed width generic unsigned integer data type.

    Generic parameters:
       N: Bit width of the :class:`Uint` [N] representation

    Args:
       val: Integer value to convert to :class:`Uint` [N]

    :class:`Uint` [N] is a generic datatype derived from :class:`Integer` [N]. It represents unsigned integers with fixed width binary representation. Concrete data type is obtained by indexing:

    >>> u16 = Uint[16]

    """
    __parameters__ = ['N']

    def __sub__(self, other):
        if (typeof(type(other), Uint)):
            res = int(self) - int(other)
            tout = type(self) - type(other)
            return tout((res, res < 0))
        else:
            return super().__sub__(other)



from pygears import gear
from pygears.sim import clk
from pygears.core.util import perpetum
import random


@gear
async def delay(din, *, f) -> b'din':
    async with din as item:
        for i in range(next(f)):
            await clk()

        yield item


def delay_rng(start, stop):
    return delay(f=perpetum(random.randint, start, stop))



import os
import logging

from pygears import PluginBase, registry
from pygears.core.hier_node import HierVisitorBase
from pygears.core.log import CustomLog
from pygears.definitions import USER_SVLIB_DIR
from pygears.svgen.intf import SVIntfGen


class SVGenInstVisitor(HierVisitorBase):
    def __init__(self):
        self.namespace = registry('SVGenModuleNamespace')
        self.svgen_map = registry('SVGenMap')

    def RTLNode(self, node):
        svgen_cls = node.params['svgen']['svgen_cls']

        if svgen_cls:
            svgen_inst = svgen_cls(node)
        else:
            svgen_inst = None

        self.svgen_map[node] = svgen_inst

    def RTLIntf(self, intf):
        self.svgen_map[intf] = SVIntfGen(intf)


def svgen_inst(top, conf):
    # if 'outdir' in conf:
    #     registry('SVGenSystemVerilogPaths').append(conf['outdir'])

    v = SVGenInstVisitor()
    v.visit(top)

    return top


def register_sv_paths(*paths):
    for p in paths:
        registry('SVGenSystemVerilogPaths').append(
            os.path.abspath(os.path.expandvars(os.path.expanduser(p))))


def svgen_log():
    return logging.getLogger('svgen')


class SVGenInstPlugin(PluginBase):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'] = {}
        cls.registry['SVGenMap'] = {}
        cls.registry['SVGenSystemVerilogPaths'] = [USER_SVLIB_DIR]
        CustomLog('svgen', logging.WARNING)

    @classmethod
    def reset(cls):
        cls.registry['SVGenMap'] = {}



from pygears import alternative, TypeMatchError, gear
from pygears.typing import Tuple
from pygears.common import fmap as common_fmap, ccat


def tuplemap_check(dtype, f):
    if not issubclass(dtype, Tuple):
        return False

    try:
        num_f = len(f)
    except TypeError as e:
        raise TypeMatchError(
            f'Tuple fmap argument "f" needs to be a sequence, received {f}')

    if len(dtype) != num_f:
        raise TypeMatchError(
            'Number of tuple types different from the number of fmap functions'
        )

    return True


@alternative(common_fmap)
@gear(enablement=b'tuplemap_check(din, f)')
def fmap(din, *, f, lvl=1, fcat=ccat, balance=None):
    lvl -= 1

    dout = []
    for i, fd in enumerate(f):
        if (lvl > 0) and (fd is not None):
            fd = common_fmap(f=fd, lvl=lvl, balance=balance)

        if fd is None:
            if balance is None:
                dout.append(din[i])
            else:
                dout.append(din[i] | balance)
        else:
            dout.append(din[i] | fd)

    return fcat(*dout)



# from pygears.typing.base import TypingMeta, type_repr
# from pygears.typing.uint import Int, Uint

# from .module_def import ModuleDefinition
# from pygears.core.module import Module
# from pygears.di.di import RequiredFeature

# from pygears.typing.match import type_match
# from pygears.typing.base import param_subs

# class SieveError(Exception):
#     pass
from pygears.core.intf import IntfOperPlugin


class PipeIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace']['__or__'] = pipe


def pipe(self, other):
    print("Here")
    return self
    # if isinstance(other, (str, TypingMeta)):
    #     return Conv(self, other, name=f'conv_{self.basename}').resolve()
    # else:
    #     return other.__ror__(self)


# def __getitem__(self, index):
#     if isinstance(index, slice):
#         start = index.start
#         if start is None:
#             start = 0

#         end = index.stop
#         if end is None:
#             end = len(self.get_type())

#         name = f'{self.proc.basename}_{start}v{end}'
#     else:
#         start = index
#         if isinstance(index, str):
#             if index not in self.get_type().fields:
#                 raise SieveError(
#                     f'No field named "{index}" in interface of type'
#                     f' {type_repr(self.get_type())}')
#             index = self.get_type().fields.index(index)

#         if index < 0:
#             start = len(self.get_type()) + index + 1
#         name = f'{self.proc.basename}_{start}'

#     return Sieve(self, index, name=name).resolve()

# def __hash__(self):
#     return id(self)

# def __eq__(self, other):
#     return ModuleDefinition(Eq, eq)(self, other)

# def __add__(self, other):
#     return ModuleDefinition(Add, add)(self, other)

# def __mul__(self, other):
#     return ModuleDefinition(Mul, mul)(self, other)

# def __truediv__(self, other):
#     return ModuleDefinition(Div, div)(self, other)

# def __sub__(self, other):
#     return ModuleDefinition(Sub, sub)(self, other)

# def __neg__(self):
#     return Neg(self, name=f'neg_{self.basename}').resolve()

# def __or__(self, other):
#     if isinstance(other, (str, TypingMeta)):
#         return Conv(self, other, name=f'conv_{self.basename}').resolve()
#     else:
#         return other.__ror__(self)

# class Sieve(Module):
#     def __init__(self, din, index, **kwds):
#         self._index = index

#         def sieve(din: '{T}') -> '{T}' + '[{}]'.format(repr(index)):
#             pass

#         super().__init__(sieve, din, **kwds)

#         # self.__call__(din, name=f'{din.basename}_{index}')

# class Oper(Module):
#     def __init__(self, func, din0, din1, **kwds):
#         from pygears.common import const
#         from pygears.util.bitw import bitw

#         if isinstance(din0, int):
#             din0 = const(tout=Uint[bitw(din0)], val=din0)

#         if isinstance(din1, int):
#             din1 = const(tout=Uint[bitw(din1)], val=din1)

#         self._din0 = din0
#         self._din1 = din1

#         # super().__init__(func, din0, din1, **kwds)

#         super().__init__(
#             func,
#             din0,
#             din1,
#             name=f'{func.__name__}_{din0.proc.basename}_{din1.proc.basename}',
#             **kwds)

#     def resolve_types(self):
#         for i, a in enumerate(self.args):
#             if not isinstance(a, Intf):
#                 print("Here")
#                 a = a().resolve()
#             self.params[f'DIN{i}_SIGNED'] = issubclass(a.type, Int)

#         return super().resolve_types()

# def eq(din0: '{TDIN0}', din1: '{TDIN1}') -> Uint[1]:
#     pass

# class Eq(Oper):
#     pass

# def add(din0: '{TDIN0}', din1: '{TDIN1}') -> '{TDIN0}+{TDIN1}':
#     pass

# class Add(Oper):
#     pass

# def sub(din0: '{TDIN0}', din1: '{TDIN1}') -> '{TDIN0}+{TDIN1}':
#     pass

# class Sub(Oper):
#     pass

# def mul(din0: '{TDIN0}', din1: '{TDIN1}') -> '{TDIN0}*{TDIN1}':
#     pass

# class Mul(Oper):
#     pass

# def div(din0: '{TDIN0}', din1: '{TDIN1}') -> '{TDIN0}/{TDIN1}':
#     pass

# class Div(Oper):
#     pass

# # class Mul(Oper):
# #     def __init__(self, din0, din1, **kwds):

# #         super().__init__(din0, din1, mul, **kwds)

# # class Div(Oper):
# #     def __init__(self, din0, din1):
# #         def div(din0: '{TDIN0}', din1: '{TDIN1}') -> '{TDIN0}*{TDIN1}':
# #             pass

# #         super().__init__(din0, din1, div)

# class Neg(Module):
#     def __init__(self, din, **kwds):
#         def neg(din: '{TDIN}') -> '{TDIN}':
#             pass

#         super().__init__(neg, din, **kwds)

# class Conv(Module):
#     def __init__(self, din, cast_type, **kwds):
#         def conv(din: '{T}') -> cast_type:
#             pass

#         super().__init__(conv, din, **kwds)

#     def resolve_types(self):
#         # print(self.args[0].get_type())
#         # print(self.name)
#         # print(self.ftypes)
#         din_type = self.args[0].get_type()
#         # self.ftypes[-1] = self.get_type()

#         if issubclass(self.get_type(),
#                       Int) and (not self.get_type().is_specified()):
#             if issubclass(din_type, Uint):
#                 self.ftypes[-1] = Int[int(din_type) + 1]
#             elif issubclass(din_type, Int):
#                 self.ftypes[-1] = din_type

#         super().resolve_types()

#         if not self.get_type().is_specified():
#             self.params.update(
#                 type_match(din_type, self.get_type(), self.params))
# self.ftypes[-1] = param_subs(self.ftypes[-1], self.params, {})



from .base import TypingMeta


class BoolMeta(TypingMeta):
    def __hash__(self):
        return 1

    def __int__(self):
        return 1


class Bool(int, metaclass=BoolMeta):
    def __new__(cls, val):
        return super(Bool, cls).__new__(cls, bool(val))

    @classmethod
    def decode(cls, val):
        return cls(val)



import operator

from pygears.core.gear import alternative, gear
from pygears.typing import Integer, Int, Uint, Bool, Tuple
from pygears.core.intf import IntfOperPlugin
from pygears.util.hof import oper_reduce
from pygears.util.utils import gather

from functools import reduce


def sub_type(dtypes):
    max_len = max(int(d) for d in dtypes)
    length = max_len + len(dtypes) - 1

    if (all(issubclass(d, Uint) for d in dtypes)):
        return Tuple[Uint[length - 1], Bool]

    return Int[length]


@gear(svgen={'svmod_fn': 'sub.sv'}, enablement=b'len(din) == 2')
async def sub(*din: Integer,
              din0_signed=b'typeof(din0, Int)',
              din1_signed=b'typeof(din1, Int)') -> b'sub_type(din)':

    async with gather(*din) as dout:
        yield reduce(operator.sub, dout)


@alternative(sub)
@gear
def sub_vararg(*din: Integer, enablement=b'len(din) > 2') -> b'sub_type(din)':
    return oper_reduce(din, sub)


class SubIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace']['__sub__'] = sub



import asyncio
import itertools
import logging
import os
import random
import sys
import tempfile
import time

from pygears import GearDone, bind, find, registry
from pygears.core.gear import GearPlugin
from pygears.core.intf import get_consumer_tree
from pygears.core.log import CustomLog, LogFmtFilter
from pygears.core.sim_event import SimEvent


def timestep():
    try:
        return registry('Timestep')
    except KeyError:
        return None


def sim_phase():
    return registry('Simulator').phase


def clk():
    registry('CurrentModule').phase = 'forward'
    return registry('ClkEvent').wait()


def delta():
    registry('CurrentModule').phase = 'back'
    return registry('DeltaEvent').wait()


def artifacts_dir():
    return registry('SimArtifactDir')


def cancel(gear):
    sim = registry('Simulator')
    sim.cancelled.add(registry('SimMap')[gear])


class SimFuture(asyncio.Future):
    # def _schedule_callbacks(self):
    #     self._loop.fut_done(self)

    def coro_iter(self):
        yield self

    def __iter__(self):
        return self.coro_iter()

    __await__ = __iter__


# A recursive function used by topo_sort
def topo_sort_util(v, g, dag, visited, stack):

    # Mark the current node as visited.
    visited[v] = True

    # Recur for all the vertices adjacent to this vertex
    for consumer in dag[g]:
        i = list(dag.keys()).index(consumer)
        if not visited[i]:
            topo_sort_util(i, consumer, dag, visited, stack)

    # Push current vertex to stack which stores result
    stack.insert(0, g)


def topo_sort(dag):
    # Mark all the vertices as not visited
    visited = [False] * len(dag)
    stack = []

    # for i, g in enumerate(dag):
    #     if not g.in_ports:
    #         stack.append(g)
    #         visited[i] = True

    # Call the recursive helper function to store Topological
    # Sort starting from all vertices one by one
    for i, g in enumerate(dag):
        if not visited[i]:
            topo_sort_util(i, g, dag, visited, stack)

    return stack


class EventLoop(asyncio.events.AbstractEventLoop):
    def __init__(self):
        self.events = {
            'before_setup': SimEvent(),
            'before_run': SimEvent(),
            'after_run': SimEvent(),
            'before_call_forward': SimEvent(),
            'after_call_forward': SimEvent(),
            'before_call_back': SimEvent(),
            'after_call_back': SimEvent(),
            'before_timestep': SimEvent(),
            'after_timestep': SimEvent(),
            'after_cleanup': SimEvent(),
            'after_cancel': SimEvent(),
            'at_exit': SimEvent()
        }

    def get_debug(self):
        return False

    def get_tasks(self):
        self.sim_map = registry('SimMap')
        dag = {}

        for g in self.sim_map:
            dag[g] = []
            for p in g.out_ports:
                dag[g].extend(
                    [port.gear for port in get_consumer_tree(p.producer)])

        gear_order = topo_sort(dag)
        # print("Topological order:")
        # for g in gear_order:
        #     print(g.name)

        # print("-"*60)

        self.sim_gears = [self.sim_map[g] for g in gear_order]
        self.tasks = {g: g.run() for g in self.sim_gears}
        self.task_data = {g: None for g in self.sim_gears}

    def call_soon(self, callback, fut):
        callback(fut)

    def fut_done(self, fut):
        sim_gear, join = self.wait_list.pop(fut)
        if fut.cancelled():
            self.cancelled.add(sim_gear)
            # print(f'Future cancelled: {sim_gear.gear.name} {join}.')
        else:
            if join:
                self.back_ready.add(sim_gear)
            else:
                self.forward_ready.add(sim_gear)

            # print(f'Future done: {sim_gear.gear.name} ready.')

    def create_future(self):
        """Create a Future object attached to the loop."""
        fut = SimFuture()
        fut.add_done_callback(self.fut_done)

        return fut

    def cancel(self, sim_gear):
        try:
            self.cur_gear = sim_gear.gear
            bind('CurrentModule', self.cur_gear)
            self.tasks[sim_gear].throw(GearDone)
        except (StopIteration, GearDone):
            pass
        else:
            sim_log().error("Gear didn't stop on cancel!")
        finally:
            self.done.add(sim_gear)
            self.events['after_cancel'](self, sim_gear)
            self.cur_gear = registry('HierRoot')
            self.back_ready.discard(sim_gear)
            self.forward_ready.discard(sim_gear)
            bind('CurrentModule', self.cur_gear)

    def run_gear(self, sim_gear, ready):
        try:
            self.cur_gear.phase = 'forward'

            if ready is self.forward_ready:
                ebef = self.events['before_call_forward']
            else:
                ebef = self.events['before_call_back']

            if ebef:
                ebef(self, sim_gear)

            data = self.tasks[sim_gear].send(self.task_data[sim_gear])

            if ready is self.forward_ready:
                eafter = self.events['after_call_forward']
            else:
                eafter = self.events['after_call_back']

            if eafter:
                eafter(self, sim_gear)
        except (StopIteration, GearDone):
            self.done.add(sim_gear)
        else:
            if isinstance(data, SimFuture):
                self.wait_list[data] = (sim_gear,
                                        self.cur_gear.phase == 'back')

            self.task_data[sim_gear] = data

    def maybe_run_gear(self, sim_gear, ready):
        if sim_gear not in ready:
            return

        ready.remove(sim_gear)

        self.cur_gear = sim_gear.gear
        bind('CurrentModule', self.cur_gear)

        self.run_gear(sim_gear, ready)

        self.cur_gear = registry('HierRoot')
        bind('CurrentModule', self.cur_gear)

    def sim_loop(self, timeout):
        clk = registry('ClkEvent')
        delta = registry('DeltaEvent')
        timestep = 0

        start_time = time.time()

        sim_log().info("-------------- Simulation start --------------")
        while self.forward_ready or self.back_ready or self.cancelled:
            self.phase = 'forward'
            for sim_gear in self.sim_gears:
                if sim_gear in self.forward_ready:
                    self.maybe_run_gear(sim_gear, self.forward_ready)

            self.phase = 'delta'
            delta.set()
            delta.clear()

            self.phase = 'back'

            for sim_gear in reversed(self.sim_gears):
                if sim_gear in self.cancelled:
                    self.cancel(sim_gear)
                    self.cancelled.remove(sim_gear)

                if sim_gear in self.back_ready:
                    self.maybe_run_gear(sim_gear, self.back_ready)

            self.phase = 'cycle'

            self.events['before_timestep'](self, timestep)

            clk.set()
            clk.clear()
            timestep += 1
            bind('Timestep', timestep)

            # if (timestep % 1000) == 0:
            #     sim_log().info("-------------- Simulation cycle --------------")

            # print(f"-------------- {timestep} ------------------")

            self.events['after_timestep'](self, timestep)
            if (timeout is not None) and (timestep == timeout):
                break

        sim_log().info(f"----------- Simulation done ---------------")
        sim_log().info(f'Elapsed: {time.time() - start_time:.2f}')

        # while self.cancelled:
        #     # print(f'Canceling {sim_gear.gear.name}')
        #     sim_gear = self.cancelled.pop()
        #     self.cancel(sim_gear)

    def run(self, timeout=None):
        self.get_tasks()
        self.wait_list = {}
        self.forward_ready = set(self.sim_gears)
        self.back_ready = set()
        self.cancelled = set()
        self.done = set()

        bind('ClkEvent', asyncio.Event())
        bind('DeltaEvent', asyncio.Event())
        bind('Timestep', 0)

        self.events['before_setup'](self)

        for sim_gear in self.sim_gears:
            self.cur_gear = sim_gear.gear
            bind('CurrentModule', self.cur_gear)
            sim_gear.setup()
            self.cur_gear = registry('HierRoot')
            bind('CurrentModule', self.cur_gear)

        self.events['before_run'](self)

        sim_exception = None
        try:
            self.sim_loop(timeout)
        except Exception as e:
            sim_exception = e

        # print(f"----------- After run ---------------")
        self.events['after_run'](self)

        if not sim_exception:
            for sim_gear in self.sim_gears:
                if sim_gear not in self.done:
                    # sim_log().debug(f"Canceling {sim_gear.gear.name}")
                    self.cancel(sim_gear)

        self.events['after_cleanup'](self)
        self.events['at_exit'](self)

        if sim_exception:
            raise sim_exception


def sim(outdir=None,
        timeout=None,
        extens=[],
        run=True,
        verbosity=logging.INFO,
        seed=None):

    if outdir is None:
        outdir = tempfile.mkdtemp()
    os.makedirs(outdir, exist_ok=True)
    bind('SimArtifactDir', outdir)

    if not seed:
        seed = int(time.time())
    random.seed(seed)
    bind('SimRandSeed', seed)
    sim_log().info(f'Running sim with seed: {seed}')

    loop = EventLoop()
    asyncio.set_event_loop(loop)
    bind('Simulator', loop)

    top = find('/')
    for oper in itertools.chain(registry('SimFlow'), extens):
        oper(top)

    if run:
        loop.run(timeout)

    return loop


class SimFmtFilter(LogFmtFilter):
    def filter(self, record):
        super().filter(record)

        m = registry('CurrentModule')

        record.module = m.name
        record.timestep = timestep()
        if record.timestep is None:
            record.timestep = '-'

        return True


class SimLog(CustomLog):
    def __init__(self, name, verbosity=logging.INFO):
        super().__init__(name, verbosity)

        # change default for error
        registry('simLog')['error']['exception'] = True
        registry('simLog')['print_traceback'] = False

    def get_format(self):
        return logging.Formatter(
            '%(timestep)s %(module)20s [%(levelname)s]: %(message)s %(err_file)s %(stack_file)s'
        )

    def get_filter(self):
        return SimFmtFilter()


def sim_log():
    return logging.getLogger('sim')


class SimPlugin(GearPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SimFlow'] = []
        cls.registry['SimTasks'] = {}
        cls.registry['SimConfig'] = {}
        cls.registry['GearExtraParams']['sim_setup'] = None
        SimLog('sim')

    @classmethod
    def reset(cls):
        bind('SimTasks', {})


def sim_assert(cond, msg=None):
    if not cond:
        sim_log().error(f'Assertion failed: {msg}')



import pydot
import os
from pygears.core.port import InPort
from pygears.core.hier_node import HierVisitorBase
from pygears.util.find import find
from pygears import registry
from pygears.util import print_hier


class Visitor(HierVisitorBase):
    def __init__(self, node_filter, outdir):
        self.gear_map = {}
        self.node_map = {}
        self.graph = pydot.Dot(
            graph_type='digraph', rankdir='LR', overlap=False)
        self.hier = [self.graph]
        self.node_filter = node_filter
        self.outdir = outdir
        os.makedirs(self.outdir, exist_ok=True)

    def enter_hier(self, module):
        self.hier.append(self.gear_map[module])

    def exit_hier(self, module):
        node = self.hier.pop()
        self.hier[-1].add_subgraph(node)

    def Gear(self, module):
        gear_fn = module.name.replace('/', '_')
        if self.outdir:
            gear_stem = os.path.abspath(os.path.join(self.outdir, gear_fn))

            v = print_hier.Visitor(params=True, fullname=True)
            v.visit(module)
            with open(f'{gear_stem}.txt', 'w') as f:
                f.write('\n'.join(v.res))

        if self.node_filter(module):
            self.gear_map[module] = pydot.Node(
                module.name,
                tooltip=module.name,
                label=module.basename,
                URL=f"localhost:5000/{gear_fn}")

            self.node_map[module] = self.gear_map[module]
            self.hier[-1].add_node(self.gear_map[module])
        else:
            self.gear_map[module] = pydot.Cluster(
                graph_name=module.name,
                label=module.basename,
                tooltip=module.name,
                fontsize=48,
                fontcolor='blue',
                labeljust='l',
                overlap=False,
                # URL=f"file://{desc_fn}")
                URL=f"localhost:5000/{gear_fn}")

            self.enter_hier(module)

            super().HierNode(module)

            self.exit_hier(module)

        return True


def _get_consumer_tree_rec(intf, consumers, node_filter):
    for port in intf.consumers:
        cons_intf = port.consumer
        if node_filter(port.gear) and (isinstance(port, InPort)):
            consumers.append(port)
        else:
            _get_consumer_tree_rec(cons_intf, consumers, node_filter)


def get_consumer_tree(intf, node_filter):
    consumers = []
    _get_consumer_tree_rec(intf, consumers, node_filter)
    return consumers


def graph(path='/', root=None, node_filter=lambda x: x, outdir=None):
    top = find(path, root)
    v = Visitor(node_filter, outdir)
    v.visit(top)

    v.edge_map = {}

    for module, node in v.node_map.items():
        for pout in module.out_ports:
            edges = get_consumer_tree(pout.producer, node_filter)

            for e in edges:
                v.edge_map[e] = pydot.Edge(
                    node,
                    v.node_map[e.gear],
                    label=f"{pout.basename} -> {e.basename}")
                v.graph.add_edge(v.edge_map[e])

    return v



from pygears import gear, alternative
from .fmap import fmap
from .flatten import flatten
from .filt import filt
from .fill import fill
from .fifo import fifo
from .mux import mux_zip, mux_valve
from .ccat import ccat
from .demux import demux_zip
from pygears.typing import Tuple, Union, Uint, typeof
from pygears.typing_common import expand as expand_type

# TODO: add implementation for arbitrary depth


@gear(enablement=b'typeof(din, Queue) and typeof(din[0], Union)')
def expand(din, *, depth=16) -> b'expand(din)':
    din_tuple = din | Tuple | fmap(f=(Tuple, None)) | flatten
    union_din = din_tuple[0, 2, 1] | expand_type(din.dtype)
    for i in range(len(union_din.dtype.types)):
        union_din = union_din | fifo(depth=depth) \
                    | fill(din=din | fifo(depth=depth) | filt(sel=i),
                           fmux=mux_zip, fdemux=demux_zip, sel=i)

    return union_din


@alternative(expand)
@gear(enablement=b'typeof(din, Tuple)')
def expand_tuple(din) -> b'expand(din)':
    ctrl_lens = []
    ctrl_list = []
    for i, t in enumerate(din.dtype):
        if(typeof(t, Union)):
            ctrl_list.append(din[i][1])
            ctrl_lens.append(int(t[1]))

    ctrl = ccat(*ctrl_list)
    ctrl = ctrl | Uint
    ctrl_width = sum(ctrl_lens)

    mux_din = []
    data_indices = []
    comb_no = 2**ctrl_width
    for i in range(comb_no):
        ctrl_bits = format(i, f'0={ctrl_width}b')
        k = 0
        for clen in ctrl_lens:
            data_indices.append(int(ctrl_bits[k:k+clen], 2))
            k += clen

        data = []
        for j, t in enumerate(din.dtype):
            if(typeof(t, Union)):
                if(data_indices[0] < len(t.types)):
                    if(int(t.types[data_indices[0]]) != 0):
                        data.append(din[j][0] | Uint[int(t.types[data_indices.pop(0)])])
                    else:
                        data_indices.pop(0)
                else:
                    data_indices.pop(0)
            else:
                data.append(din[j])
        mux_din.append(ccat(*data))

    return (ctrl, *mux_din) | mux_valve | expand_type(din.dtype)


@alternative(expand)
@gear(enablement=b'typeof(din, Queue) and typeof(din[0], Tuple)')
def expand_queue_to_tuple(din) -> b'expand(din)':
    din_tuple = din | Tuple | flatten
    tuple_len = len(din_tuple.dtype)

    indices = []
    for i in range(tuple_len-1):
        indices.append(i)
        indices.append(tuple_len-1)

    return din_tuple[tuple(indices)] | expand_type(din.dtype)



from pygears import registry


def _find_rec(path, root):
    parts = path.split("/")

    if parts[0] == '..':
        return _find_rec("/".join(parts[1:]), root.parent)

    for c in root.child:
        if hasattr(c, 'basename') and c.basename == parts[0]:
            child = c
            break
    else:
        raise ModuleNotFoundError()

    if len(parts) == 1:
        return child
    else:
        return _find_rec("/".join(parts[1:]), child)


def find(path, root=None):
    if path.startswith('/'):
        path = path[1:]
        root = registry('HierRoot')
    else:
        if path.startswith('./'):
            path = path[2:]

        root = registry('CurrentModule')

    if path == '':
        return root

    try:
        return _find_rec(path, root)
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f'No module found on path "{path}"')



from pygears import alternative, gear
from pygears.typing import Queue
from pygears.common import quenvelope, cart
from pygears.common import fmap as common_fmap


@alternative(common_fmap)
@gear(enablement=b'issubclass(din, Queue)')
def fmap(din, *, f, lvl=1, fcat=cart, balance=None, common_balance=True):
    queue_lvl = din.dtype.lvl
    fmap_lvl = min(lvl, queue_lvl)
    lvl -= fmap_lvl

    env = din | quenvelope(lvl=fmap_lvl)
    data = din[0:queue_lvl - fmap_lvl + 1]

    if lvl > 0:
        f = common_fmap(f=f, lvl=lvl, fcat=fcat, balance=balance)

    dout = data | f

    if not isinstance(dout, tuple):
        dout = (dout, )

    if balance is not None and common_balance:
        env = env | balance

    cat_dout = []
    for d in dout:
        benv = env
        if balance is not None and not common_balance:
            benv = env | balance

        cat_dout.append(fcat(benv, d))

    # Cast to remove tuple Unit from resulting cart tuple
    ret = [d | Queue[d.dtype[0][1], d.dtype.lvl] for d in cat_dout]

    return tuple(ret)



from pygears.sim.scv import SCVTypeSeqVisitor, scv_compile
from pygears.sim import artifacts_dir
from pygears import module
import os


def dtype_rnd_seq(t, outdir=None, cons=None):
    if cons is not None:
        if not outdir:
            outdir = os.path.join(artifacts_dir(), module().basename)

        scvlib = scv_compile(outdir, module().basename, cons)
        scvlib.randomize_seed()
    else:
        scvlib = None

    yield SCVTypeSeqVisitor(scvlib).visit(t)



from pygears.core.hier_node import HierVisitorBase, HierYielderBase
import os
import jinja2
from pygears.util.fileio import save_file
from pygears import registry
from pygears.svgen.util import svgen_typedef
from pygears.rtl.inst import RTLNodeDesign
from pygears.typing import bitw


def format_list(list_, pattern):
    return [pattern % s for s in list_]


def keymap(list_, key):
    return [item[key] for item in list_]


def isinput(list_):
    return [item for item in list_ if item["modport"] == "consumer"]


def isoutput(list_):
    return [item for item in list_ if item["modport"] == "producer"]


def startswith(field, s):
    return field.startswith(s)


class TemplateEnv:
    def __init__(self):
        self.basedir = os.path.dirname(__file__)
        self.templates = {}
        self.jenv = jinja2.Environment(
            extensions=['jinja2.ext.do'], trim_blocks=True, lstrip_blocks=True)
        self.jenv.globals.update(
            zip=zip,
            len=len,
            int=int,
            bitw=bitw,
            enumerate=enumerate,
            svgen_typedef=svgen_typedef)

        self.jenv.filters['format_list'] = format_list
        self.jenv.filters['keymap'] = keymap
        self.jenv.filters['isinput'] = isinput
        self.jenv.filters['isoutput'] = isoutput
        self.jenv.tests['startswith'] = startswith

        self.snippets = self.load(self.basedir, 'snippet.j2').module

    def load(self, tmplt_dir, tmplt_fn):
        key = os.path.join(self.basedir, tmplt_dir, tmplt_fn)
        if key not in self.templates:
            self.jenv.loader = jinja2.FileSystemLoader([self.basedir, tmplt_dir])
            template = self.jenv.get_template(tmplt_fn)
            self.templates[key] = template

        return self.templates[key]

    def render_local(self, fn, tmplt_fn, context):
        return self.render(os.path.dirname(fn), tmplt_fn, context)

    def render(self, tmplt_dir, tmplt_fn, context):
        return self.load(tmplt_dir, tmplt_fn).render(context)


class SVGenGenerateVisitor(HierYielderBase):
    def __init__(self, top, wrapper=False):
        self.template_env = TemplateEnv()
        self.svgen_map = registry('SVGenMap')
        self.wrapper = wrapper
        self.top = top

    def RTLNode(self, node):
        svgen = self.svgen_map.get(node, None)
        if svgen is not None:
            contents = svgen.get_module(self.template_env)
            # print(f'Generating {svgen.sv_file_name}')
            yield svgen.sv_file_name, contents

            if (self.wrapper) and (node == self.top):
                yield f'wrap_{svgen.sv_file_name}', svgen.get_synth_wrap(
                    self.template_env)


def svgen_module(node):
    v = SVGenGenerateVisitor()
    svgen, contents = next(v.visit(node))
    return contents


def svgen_yield(top):
    v = SVGenGenerateVisitor()
    for svgen, contents in v.visit(top):
        if contents:
            yield svgen, contents


def svgen_generate(top, conf):
    v = SVGenGenerateVisitor(top, conf.get('wrapper', False))
    for file_names, contents in v.visit(top):
        if contents:
            if isinstance(contents, (tuple, list)):
                for fn, c in zip(file_names, contents):
                    save_file(fn, conf['outdir'], c)
            else:
                save_file(file_names, conf['outdir'], contents)

    return top



import copy
import inspect
import functools
import asyncio
import sys

from pygears.registry import PluginBase, bind, registry
from pygears.typing import Any

from .hier_node import NamedHierNode
from .infer_ftypes import TypeMatchError, infer_ftypes, type_is_specified
from .intf import Intf
from .partial import Partial
from .port import InPort, OutPort
from .util import doublewrap
from .log import core_log

code_map = {}


class TooManyArguments(Exception):
    pass


class GearTypeNotSpecified(Exception):
    pass


class GearArgsNotSpecified(Exception):
    pass


def check_arg_num(argnames, varargsname, args):
    if (len(args) < len(argnames)) or (not varargsname and
                                       (len(args) > len(argnames))):
        balance = "few" if (len(args) < len(argnames)) else "many"

        raise TooManyArguments(f"Too {balance} arguments provided.")


def check_arg_specified(args):
    args_res = []
    const_args_gears = []
    for i, a in enumerate(args):
        if isinstance(a, Partial):
            raise GearArgsNotSpecified(f"Unresolved input arg {i}")

        if not isinstance(a, Intf):
            from pygears.common import const
            try:
                a = const(val=a)
                const_args_gears.append(module().child[-1])
            except GearTypeNotSpecified:
                raise GearArgsNotSpecified(f"Unresolved input arg {i}")

        args_res.append(a)

        if not type_is_specified(a.dtype):
            raise GearArgsNotSpecified(
                f"Input arg {i} has unresolved type {repr(a.dtype)}")

    return tuple(args_res), const_args_gears


class create_hier:
    def __init__(self, gear):
        self.gear = gear

    def __enter__(self):
        bind('CurrentModule', self.gear)
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        bind('CurrentModule', self.gear.parent)
        if exception_type is not None:
            self.gear.clear()


class Gear(NamedHierNode):
    def __new__(cls, func, meta_kwds, *args, name=None, __base__=None, **kwds):

        if name is None:
            if __base__ is None:
                name = func.__name__
            else:
                name = __base__.__name__

        kwds_comb = kwds.copy()
        kwds_comb.update(meta_kwds)

        gear = super().__new__(cls)
        try:
            gear.__init__(func, *args, name=name, **kwds_comb)
        except Exception as e:
            gear.remove()
            raise e

        if not gear.params.pop('enablement'):
            gear.remove()
            raise TypeMatchError(
                f'Enablement condition failed for "{gear.name}" alternative'
                f' "{gear.definition.__module__}.{gear.definition.__name__}": '
                f'{meta_kwds["enablement"]}')

        return gear.resolve()

    def __init__(self, func, *args, name=None, intfs=None, outnames=[], **kwds):
        super().__init__(name, registry('CurrentModule'))

        self.in_ports = []
        self.out_ports = []
        self.const_args_gears = []

        self.func = func
        self.__doc__ = func.__doc__

        self.outnames = outnames.copy()
        if intfs is None:
            self.fix_intfs = []
        elif isinstance(intfs, Intf):
            self.fix_intfs = [intfs]
        else:
            self.fix_intfs = intfs.copy()

        self.args = args
        self.resolved = False

        argspec = inspect.getfullargspec(func)
        self.argnames = argspec.args
        self.varargsname = argspec.varargs
        self.annotations = argspec.annotations
        self.kwdnames = argspec.kwonlyargs

        try:
            check_arg_num(self.argnames, self.varargsname, self.args)
        except TooManyArguments as e:
            TooManyArguments(f'{e}, for the module {self.name}')

        try:
            self.args, self.const_args_gears = check_arg_specified(self.args)
        except GearArgsNotSpecified as e:
            raise GearArgsNotSpecified(
                f'{str(e)}, when instantiating {self.name}')

        self.params = {}
        if isinstance(argspec.kwonlydefaults, dict):
            self.params.update(argspec.kwonlydefaults)

        self.params.update(kwds)

        self.params.update({
            a: (self.annotations[a] if a in self.annotations else Any)
            for a in self.argnames
        })

        self._handle_return_annot()
        self._expand_varargs()
        self.in_ports = [
            InPort(self, i, name) for i, name in enumerate(self.argnames)
        ]

        for i, a in enumerate(self.args):
            try:
                a.connect(self.in_ports[i])
            except AttributeError:
                raise GearArgsNotSpecified(
                    f"Input arg {i} for module {self.name} was not"
                    f" resolved to interface, instead {repr(a)} received")

        self.infer_params()

    def _handle_return_annot(self):
        if "return" in self.annotations:
            ret_anot = self.annotations["return"]
            if isinstance(ret_anot, dict):
                self.outnames = tuple(ret_anot.keys())
                self.params['return'] = tuple(ret_anot.values())
            else:
                self.params['return'] = ret_anot
        else:
            self.params['return'] = None

    def _expand_varargs(self):
        if self.varargsname:
            vararg_type_list = []
            if self.varargsname in self.annotations:
                vararg_type = self.annotations[self.varargsname]
            else:
                vararg_type = Any
            # Append the types of the self.varargsname
            for i, a in enumerate(self.args[len(self.argnames):]):
                if isinstance(vararg_type, str):
                    # If vararg_type is a template string, it can be made
                    # dependent on the arguments position
                    type_tmpl_i = vararg_type.format(i).encode()
                else:
                    # Vararg is not a template and should be passed as is
                    type_tmpl_i = vararg_type

                argname = f'{self.varargsname}{i}'

                vararg_type_list.append(argname)
                self.params[argname] = type_tmpl_i
                self.argnames.append(argname)

            self.params[
                self.
                varargsname] = f'({", ".join(vararg_type_list)}, )'.encode()

    def remove(self):
        for p in self.in_ports:
            if p.producer is not None:
                try:
                    p.producer.disconnect(p)
                except ValueError:
                    pass

        for p in self.out_ports:
            if p.producer is not None:
                p.producer.disconnect(p)

        for g in self.const_args_gears:
            g.remove()

        try:
            super().remove()
        except ValueError:
            pass

    @property
    def definition(self):
        return self.params['definition']

    @property
    def dout(self):
        if len(self.intfs) > 1:
            return tuple(p.producer for p in self.out_ports)
        else:
            return self.out_ports[0].producer

    @property
    def tout(self):
        if len(self.intfs) > 1:
            return tuple(i.dtype for i in self.intfs)
        else:
            return self.intfs[0].dtype

    def set_ftype(self, ft, i):
        self.dtype_templates[i] = ft

    def is_specified(self):
        for i in self.intfs:
            if not type_is_specified(i.dtype):
                return False
        else:
            return True

    def get_arg_types(self):
        return tuple(a.dtype for a in self.args)

    def get_type(self):
        if len(self.intfs) > 1:
            return tuple(i.dtype for i in self.intfs)
        elif len(self.intfs) == 1:
            return self.intfs[0].dtype
        else:
            return None

    def infer_params(self):
        arg_types = {
            name: arg.dtype
            for name, arg in zip(self.argnames, self.args)
        }

        try:
            self.params = infer_ftypes(
                self.params,
                arg_types,
                namespace=self.func.__globals__,
                allow_incomplete=False)
        except TypeMatchError as e:
            raise TypeMatchError(f'{str(e)}, of the module "{self.name}"')

    def resolve(self):
        for port in self.in_ports:
            Intf(port.dtype).source(port)

        is_async_gen = bool(
            self.func.__code__.co_flags & inspect.CO_ASYNC_GENERATOR)
        func_ret = tuple()
        if (not inspect.iscoroutinefunction(self.func)
                and not inspect.isgeneratorfunction(self.func)
                and not is_async_gen):
            func_ret = self.resolve_func()

        out_dtype = tuple()
        if func_ret:
            out_dtype = tuple(r.dtype for r in func_ret)
        elif self.params['return'] is not None:
            if not isinstance(self.params['return'], tuple):
                out_dtype = (self.params['return'], )
            else:
                out_dtype = self.params['return']

        if (len(self.outnames) == 0) and (len(out_dtype) == 1):
            self.outnames.append('dout')
        else:
            for i in range(len(self.outnames), len(out_dtype)):
                self.outnames.append(f'dout{i}')

        self.out_ports = [
            OutPort(self, i, name) for i, name in enumerate(self.outnames)
        ]

        # Connect internal interfaces
        if func_ret:
            for i, r in enumerate(func_ret):
                r.connect(self.out_ports[i])
        else:
            for dtype, port in zip(out_dtype, self.out_ports):
                Intf(dtype).connect(port)

        # Connect output interfaces
        self.intfs = []
        out_intfs = []
        if isinstance(self.fix_intfs, dict):
            for i, (name, dt) in enumerate(zip(self.outnames, out_dtype)):
                if name in self.fix_intfs:
                    intf = self.fix_intfs[name]
                else:
                    intf = Intf(dt)
                    out_intfs.append(intf)

                self.intfs.append(intf)

        elif self.fix_intfs:
            self.intfs = self.fix_intfs
        else:
            self.intfs = [Intf(dt) for dt in out_dtype]
            out_intfs = self.intfs

        assert len(self.intfs) == len(out_dtype)
        for intf, port in zip(self.intfs, self.out_ports):
            intf.source(port)

        for name, dtype in zip(self.outnames, out_dtype):
            self.params[name] = dtype

        if not self.is_specified():
            raise GearTypeNotSpecified(
                f"Output type of the module {self.name}"
                f" could not be resolved, and resulted in {repr(out_dtype)}")

        for c in self.child:
            for p in c.out_ports:
                intf = p.consumer
                if intf not in self.intfs and not intf.consumers:
                    core_log().warning(f'{c.name}.{p.basename} left dangling.')

        if len(out_intfs) > 1:
            return tuple(out_intfs)
        elif len(out_intfs) == 1:
            return out_intfs[0]
        else:
            return None

    def resolve_func(self):
        with create_hier(self):
            func_args = [p.consumer for p in self.in_ports]

            func_kwds = {
                k: self.params[k]
                for k in self.kwdnames if k in self.params
            }

            self.func_locals = {}
            code_map = registry('GearCodeMap')
            code_map[self.func.__code__] = self

            def tracer(frame, event, arg):
                if event == 'return':
                    if frame.f_code in code_map:
                        code_map[
                            frame.f_code].func_locals = frame.f_locals.copy()

            # tracer is activated on next call, return or exception
            if registry('CurrentModule').parent == registry('HierRoot'):
                sys.setprofile(tracer)

            ret = self.func(*func_args, **func_kwds)

            if registry('CurrentModule').parent == registry('HierRoot'):
                sys.setprofile(None)

            for name, val in self.func_locals.items():
                if isinstance(val, Intf):
                    val.var_name = name

        # if not any([isinstance(c, Gear) for c in self.child]):
        #     self.clear()

        if ret is None:
            ret = tuple()
        elif not isinstance(ret, tuple):
            ret = (ret, )

        return ret


def alternative(*base_gear_defs):
    def gear_decorator(gear_def):
        for d in base_gear_defs:
            alternatives = getattr(d.func, 'alternatives', [])
            alternatives.append(gear_def.func)
            d.func.alternatives = alternatives
        return gear_def

    return gear_decorator


@doublewrap
def gear(func, gear_cls=Gear, **meta_kwds):
    from pygears.core.funcutils import FunctionBuilder
    fb = FunctionBuilder.from_func(func)
    fb.filename = '<string>'

    # Add defaults from GearExtraParams registry
    for k, v in registry('GearExtraParams').items():
        if k not in fb.kwonlyargs:
            fb.kwonlyargs.append(k)
            fb.kwonlydefaults[k] = copy.copy(v)

    fb.body = (f"return gear_cls(gear_func, meta_kwds, "
               f"{fb.get_invocation_str()})")

    # Add defaults from GearMetaParams registry
    for k, v in registry('GearMetaParams').items():
        if k not in meta_kwds:
            meta_kwds[k] = copy.copy(v)

    execdict = {
        'gear_cls': gear_cls,
        'meta_kwds': meta_kwds,
        'gear_func': func
    }
    execdict.update(func.__globals__)
    execdict_keys = list(execdict.keys())
    execdict_values = list(execdict.values())

    def formatannotation(annotation, base_module=None):
        try:
            return execdict_keys[execdict_values.index(annotation)]
        except ValueError:
            if not isinstance(str, bytes):
                return '"b' + repr(annotation) + '"'
            else:
                return annotation

    gear_func = fb.get_func(
        execdict=execdict, formatannotation=formatannotation)

    functools.update_wrapper(gear_func, func)

    p = Partial(gear_func)
    meta_kwds['definition'] = p
    p.meta_kwds = meta_kwds

    return p


def module():
    return registry('CurrentModule')


class GearPlugin(PluginBase):
    @classmethod
    def bind(cls):
        cls.registry['HierRoot'] = NamedHierNode('')
        cls.registry['CurrentModule'] = cls.registry['HierRoot']
        cls.registry['GearCodeMap'] = {}
        cls.registry['GearMetaParams'] = {'enablement': True}
        cls.registry['GearExtraParams'] = {
            'name': None,
            'intfs': [],
            'outnames': [],
            '__base__': None
        }

    @classmethod
    def reset(cls):
        bind('HierRoot', NamedHierNode(''))
        bind('CurrentModule', cls.registry['HierRoot'])
        bind('GearCodeMap', {})



from pygears import gear
from pygears.typing import Tuple, Uint

TWrDin = Tuple[Uint['w_addr'], 'w_data']
TRdDin = Uint['w_addr']

outnames = ['rd_data_if']


@gear(outnames=outnames)
def sdp_rd_port(rd_addr_if: TRdDin, *, w_data, depth) -> b'w_data':
    pass


@gear
def sdp_wr_port(wr_addr_data_if: TWrDin) -> None:
    pass


@gear(
    outnames=outnames, sv_submodules=['sdp_mem', 'sdp_rd_port', 'sdp_wr_port'])
def sdp(wr_addr_data_if: TWrDin,
        rd_addr_if: TRdDin,
        *,
        w_data=b'w_data',
        w_addr=b'w_addr',
        depth=5) -> b'w_data':
    pass



import pprint
from pygears import registry
from pygears.core.hier_node import HierVisitorBase
import textwrap
from pygears.core.log import util_log


def module_signature(module, fullname, params):
    t = module.get_type()
    if t is None:
        types = "None"
        sizes = ""
    elif isinstance(t, tuple):
        types = ', '.join([str(tt) for tt in t])
        sizes = ', '.join([str(int(tt)) for tt in t])
    else:
        types = str(t)
        sizes = int(t)

    if fullname:
        name = module.name
    else:
        name = module.basename

    return f'{name}: {types} ({sizes})'


class Visitor(HierVisitorBase):
    omitted = ['definition', 'sim_setup']

    def __init__(self, params=False, fullname=False):
        self.indent = ""
        self.params = params
        self.fullname = fullname
        self.res = []
        self.pp = pprint.PrettyPrinter(indent=4, width=120)

    def Gear(self, node):
        self.print_module_signature(node)
        self.indent += "    "
        super().HierNode(node)
        self.indent = self.indent[:-4]
        return True

    def print_module_signature(self, module):
        def get_size(t):
            try:
                return str(int(t))
            except TypeError:
                return '?'

        t = module.get_type()
        if t is None:
            types = "None"
            sizes = ""
        elif isinstance(t, tuple):
            types = ', '.join([str(tt) for tt in t])
            sizes = ', '.join([get_size(tt) for tt in t])
        else:
            types = str(t)
            sizes = get_size(t)

        if self.fullname:
            name = module.name
        else:
            name = module.basename

        self.res.append(f'{self.indent}{name}: {types} ({sizes})')
        if self.params:
            params = {
                p: repr(v)
                for p, v in module.params.items() if p not in self.omitted
            }

            p = self.pp.pformat(params)
            # print(p)
            self.res.append(textwrap.indent(p, self.indent))
            # self.res.append(f'{self.indent}    {p}')


def print_hier(root=None, params=False, fullname=False):
    if root is None:
        root = registry('HierRoot')

    v = Visitor(params, fullname)
    v.visit(root)
    util_log().info('\n'.join(v.res))



from pygears import gear


@gear(enablement=b'len(din) == 2')
async def max2(*din,
               din0_signed=b'issubclass(din0, Int)',
               din1_signed=b'issubclass(din1, Int)') -> b'din[0]':
    pass



from pygears import gear, alternative
from pygears.common.fmap import fmap
from pygears.common.flatten import flatten
from pygears.typing import Tuple, typeof, Queue, Union, Uint
from pygears.typing_common import factor as factor_type


def is_union_of_queues_with_equal_len(din):
    lengths = list(map(int, din.types))
    return typeof(din, Union) and \
        all(typeof(t, Queue) for t in din.types) and \
        lengths[1:] == lengths[:-1]


# TODO: consider what to do with eots
@gear(enablement=b'is_union_of_queues_with_equal_len(din)')
def factor(din) -> b'factor(din)':
    lvl = din.dtype.types[0].lvl
    data_len = int(din.dtype.types[0]) - lvl
    din_tuple = din | Tuple \
        | fmap(f=(Tuple[Uint[data_len], Uint[lvl]], None)) \
        | flatten

    return din_tuple[0, 2, 1] | factor_type(din.dtype)


# TODO: consider what to do with eots
@alternative(factor)
@gear(enablement=b'typeof(din, Tuple) and all(typeof(t, Queue) for t in din)')
def factor_tuple_queue(din) -> b'factor(din)':
    din_tuple = din | fmap(f=(Tuple, )*len(din.dtype)) | flatten

    tuple_len = len(din_tuple.dtype)
    indices = list(range(0, tuple_len, 2))
    indices.append(1)  # eot

    return din_tuple[tuple(indices)] | factor_type(din.dtype)



from pygears import gear
from pygears.typing import Queue, Uint


@gear(svgen={'svmod_fn': 'qlen_cnt.sv'})
async def qlen_cnt(din: Queue['tdin', 'din_lvl'],
                   *,
                   cnt_lvl=1,
                   cnt_one_more=False,
                   w_out=16) -> Uint['w_out']:
    '''Outputs only one value when input eots'''
    pass



from pygears import gear
from pygears.svgen.inst import SVGenInstPlugin
from pygears.svgen.svmod import SVModuleGen
from pygears.typing import Queue, Uint, Tuple


def chunk_type(dtypes, chunk_size):
    if (chunk_size == 1):
        return Queue[dtypes[0]]
    else:
        return Queue[Tuple[(dtypes[0][0], ) * chunk_size], 2]


@gear
def chunk_concat(cfg: Uint['Tn'],
                 *din: 'w_din{0}',
                 cnt_type=0,
                 chunk_size=1,
                 pad=0) -> b'chunk_type(din, chunk_size)':
    pass


class SVGenChunkConcat(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        context = {
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs()),
            'cnt_type': self.node.params['cnt_type'],
            'chunk_size': self.node.params['chunk_size'],
            'pad': self.node.params['pad']
        }
        return template_env.render_local(__file__, "chunk_concat.j2", context)


class SVGenChunkConcatPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][chunk_concat] = SVGenChunkConcat



from pygears import gear
from pygears.typing import Union
from pygears.common.shred import shred
from pygears.common.ccat import ccat
from pygears.common.fmap import fmap
from pygears.common.filt import filt


@gear(enablement=b'len(din) == 2')
def union_sync(*din, ctrl, outsync=True) -> b'din':
    pass


@gear
def union_sync_with(din, sync_in, *, ctrl, balance=None):
    if balance:
        sync_in = sync_in | balance

    sync_in_sync, din_sync = union_sync(sync_in, din, ctrl=ctrl)
    sync_in_sync | shred

    return din_sync


@gear
def filt_by(din, ctrl, *, sel, fcat=ccat):
    return fcat(din, ctrl) \
        | Union \
        | filt(sel=sel)


@gear
def pick_by(din, ctrl, *, f, fcat=ccat, **kwds):
    try:
        len(f)
    except TypeError:
        f = (None, f)

    return fcat(din, ctrl) \
        | Union \
        | fmap(f=f, **kwds) \
        | union_collapse


@gear
def do_if(cond, din, *, f, fe=None, fcat=ccat, **kwds):
    return fcat(din, cond) \
        | Union \
        | fmap(f=(fe, f), **kwds) \
        | union_collapse


def all_same(din):
    return din.types.count(din.types[0]) == len(din.types)


@gear(enablement=b'all_same(din)')
def union_collapse(din: Union) -> b'din.types[0]':
    return din[0] | din.dtype.types[0]



from pygears import gear
from pygears.sim import clk
from pygears.sim.utils import SimDelay
from pygears.typing import TLM


class Partial:
    def __new__(cls, val):
        if isinstance(val, Partial):
            return val
        else:
            obj = super().__new__(cls)
            obj.__init__(val)
            return obj

    def __init__(self, val):
        self._val = val

    @property
    def val(self):
        return self._val


class TypeMonitorVisitor:
    def __init__(self, dtype):
        self.data = None
        self.dtype = dtype

    def __bool__(self):
        return isinstance(self.data, Partial)

    def append(self, elem):
        self.data = self.visit(self.data, elem, self.dtype)
        return self.data

    def visit(self, data, elem, dtype):
        visit_func_name = f'visit_{dtype.__name__.lower()}'

        visit_func = getattr(self, visit_func_name, self.visit_default)

        return visit_func(data, elem, dtype)

    def visit_default(self, data, elem, dtype):
        return elem

    def visit_queue(self, data, elem, dtype):
        if dtype.lvl == 1:
            sub_elem = elem[0]
        else:
            sub_elem = elem[:-1]

        if not data:
            sub_data = None
            data = []
        else:
            data = data.val
            if isinstance(data[-1], Partial):
                sub_data = data.pop()
            else:
                sub_data = None

        sub_data = self.visit(sub_data, sub_elem, dtype[:-1])
        data.append(sub_data)

        eot = elem[-1]
        if eot and (not isinstance(sub_data, Partial)):
            return data
        else:
            return Partial(data)


@gear
async def mon(din, *, t=b'din') -> TLM['din']:
    v = TypeMonitorVisitor(t)
    while 1:
        data = None
        while (isinstance(data, Partial) or data is None):
            # print('Monitor waiting')
            item = await din.get()
            # print('Monitor got: ', item)
            data = v.visit(data, item, t)

        # print('Monitor emits: ', data)
        yield data


@gear
async def delay_mon(din, *, t=b'din', delay=SimDelay(0, 0)) -> b'din':
    async with din as item:
        await delay.delay
        yield item



import fnmatch
from string import Template

from pygears import PluginBase, registry
from pygears.rtl.port import InPort, OutPort
from pygears.svgen.util import svgen_typedef

spy_connect_t = Template("""
dti_spy #(${intf_name}_t) _${intf_name}(clk, rst);
assign _${intf_name}.data = ${conn_name}.data;
assign _${intf_name}.valid = ${conn_name}.valid;
assign _${intf_name}.ready = ${conn_name}.ready;""")


class SVIntfGen:
    def __init__(self, intf):
        self.intf = intf

    @property
    def basename(self):
        return self.intf.basename

    @property
    def outname(self):
        return self.intf.outname

    def get_inst(self, template_env):
        if self.intf.producer is None:
            return

        inst = []
        if not self.intf.is_port_intf:
            inst.append(self.get_intf_def(self.basename, 1, template_env))

        gen_dbg_intf = any(
            fnmatch.fnmatch(self.intf.parent.gear.name, p)
            for p in registry('SVGenDebugIntfs'))

        if self.intf.is_broadcast:
            inst.extend([
                self.get_intf_def(self.outname, len(self.intf.consumers),
                                  template_env),
                self.get_bc_module(template_env)
            ])

            for i, cons_port in enumerate(self.intf.consumers):
                if isinstance(cons_port, OutPort):
                    inst.append(self.get_connect_module(i, template_env))

            if gen_dbg_intf:
                for i in range(len(self.intf.consumers)):
                    intf_name = f'{self.outname}_{i}'
                    conn_name = f'{self.outname}[{i}]'
                    inst.extend(
                        svgen_typedef(self.intf.dtype, intf_name).split('\n'))

                    inst.extend(
                        spy_connect_t.substitute(
                            intf_name=intf_name,
                            conn_name=conn_name).split('\n'))

        if gen_dbg_intf:
            inst.extend(
                svgen_typedef(self.intf.dtype, self.basename).split('\n'))

            inst.extend(
                spy_connect_t.substitute(
                    intf_name=self.basename,
                    conn_name=self.basename).split('\n'))

        return '\n'.join(inst)

    def get_intf_def(self, name, size, template_env):
        ctx = {
            'name': name,
            'width': int(self.intf.dtype),
            'size': size,
            'type': str(self.intf.dtype)
        }

        return template_env.snippets.intf_inst(**ctx)

    def get_connect_module(self, index, template_env):
        inst_name = f'connect_{self.basename}'
        din_name = self.outname
        if self.intf.is_broadcast:
            inst_name += f'_{index}'
            din_name += f'[{index}]'

        connect_context = {
            'module_name': 'connect',
            'inst_name': inst_name,
            'param_map': {},
            'port_map': {
                'din': din_name,
                'dout': self.intf.consumers[index].basename
            }
        }

        return template_env.snippets.module_inst(**connect_context)

    def get_bc_module(self, template_env):
        inst_name = f'bc_{self.basename}'
        if inst_name.endswith('_if_s'):
            inst_name = inst_name[:-len('_if_s')]

        rst_name = 'local_rst' if registry("SVGenMap")[
            self.intf.parent].has_local_rst else 'rst'

        bc_context = {
            'rst_name': rst_name,
            'module_name': 'bc',
            'inst_name': inst_name,
            'param_map': {
                'SIZE': len(self.intf.consumers)
            },
            'port_map': {
                'din': self.basename,
                'dout': self.outname
            }
        }

        return template_env.snippets.module_inst(**bc_context)


class SVGenIntfPlugin(PluginBase):
    @classmethod
    def bind(cls):
        cls.registry['SVGenDebugIntfs'] = []



from pygears import gear
from pygears.core.log import gear_log
from pygears.svgen.inst import SVGenInstPlugin
from pygears.svgen.svmod import SVModuleGen
from pygears.typing import Queue


def trr_dist_type(dtype):
    return Queue[dtype[0]]


@gear
async def trr_dist(din: Queue, *,
                   dout_num) -> b'(trr_dist_type(din), ) * dout_num':
    t_din = din.dtype

    for i in range(dout_num):
        out_res = [None] * dout_num
        val = t_din((0, 0, 0))

        while (val.eot[0] == 0):
            async with din as val:
                out_res[i] = val[:-1]
                gear_log().debug(
                    f'Trr_dist yielding on output {i} value {out_res[i]}')
                yield tuple(out_res)

        if val.eot == int('1' * t_din.lvl, 2):
            gear_log().debug(f'Trr_dist reset to first output')
            break


class SVGenTrrDist(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        context = {
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs())
        }
        return template_env.render_local(__file__, "trr_dist.j2", context)


class SVGenTrrDistPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][trr_dist] = SVGenTrrDist



from pygears import gear, registry
from pygears.typing import Union
from pygears.common.demux import demux
from pygears.common.mux import mux
from pygears.common.shred import shred


def fill_type(din_t, union_t, sel):
    dtypes = union_t.types.copy()
    dtypes[sel] = din_t

    return Union[tuple(dtypes)]


@gear
def fill(din,
         union_din: Union,
         *,
         fdemux=demux(ctrl_out=True),
         fmux=mux,
         sel) -> b'fill_type(din, union_din, sel)':
    fields = union_din | fdemux
    fields_list = list(fields)

    fields_list[sel+1] | shred

    fields_list[sel+1] = din
    return tuple(fields_list) | fmux



from pygears.core.gear import alternative, gear
from pygears.typing import Integer, Int, Uint
from pygears.core.intf import IntfOperPlugin


@gear(svgen={'svmod_fn': 'invert.sv'})
def invert(din) -> b'din':
    pass


class InvertIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace']['__invert__'] = invert



from pygears.typing.base import TypingMeta


class UnitMeta(TypingMeta):
    def __int__(self):
        return 0

    def __str__(self):
        return '()'


class Unit(metaclass=UnitMeta):
    def __init__(self, v=None):
        if (v is not None) and (not isinstance(v, Unit)):
            raise TypeError

    def __str__(self):
        return '()'

    def __repr__(self):
        return 'Unit()'

    def __eq__(self, other):
        return isinstance(other, Unit)



# ===========================================================================
# Logger usage:
#
# To add new logger:
#  - create CustomLog instance
#  - create <name>_log() wrapper function
#
# To overwrite default behaviour:
#  - change appropriate values in registry(<name>Log)[verbosity][action]
#    - verbosity: error or warning
#    - action: exception or debug
#    For example: registry('coreLog')['error']['exception'] = False
#  - to print traceback on fail: registry(<name>Log)['print_traceback'] = True
#
# To change verbosity of displayed messages set the desired level:
#   For example: core_log().level = logging.ERROR
#
# ===========================================================================

import copy
import logging
import sys
import tempfile
from string import Template

from pygears import bind, registry
from pygears.registry import PluginBase

from .err import enum_stacktrace

registry_log_name = Template('${name}Log')


class LogException(Exception):
    def __init__(self, message, name):
        super().__init__(message)
        self.name = name


class LogWrap:
    def __init__(self, logger):
        self.logger = logger
        self.name = logger.name

    def severity_action(self, severity, message):
        log_cfg = registry(registry_log_name.substitute(name=self.name))
        if log_cfg[severity]['exception']:
            raise LogException(message, self.name)
        elif log_cfg[severity]['debug']:
            import pdb
            pdb.set_trace()

    def stack_trace(self, verbosity, message):
        stack_traceback_fn = registry('StackTracebackFn')
        with open(stack_traceback_fn, 'a') as f:
            delim = '-' * 50 + '\n'
            f.write(delim)
            f.write(f'{self.name} [{verbosity.upper()}] {message}\n\n')
            tr = ''
            exc_type, exc_value, exc_traceback = sys.exc_info()
            for s in enum_stacktrace():
                tr += s
            f.write(tr)
            f.write(delim)

    def warning(self, message, *args, **kws):
        if self.logger.isEnabledFor(logging.WARNING):
            severity = 'warning'
            self.logger._log(logging.WARNING, message, args, **kws)
            self.stack_trace(severity, message)
            self.severity_action(severity, message)

    def error(self, message, *args, **kws):
        if self.logger.isEnabledFor(logging.ERROR):
            severity = 'error'
            self.logger._log(logging.ERROR, message, args, **kws)
            self.stack_trace(severity, message)
            self.severity_action(severity, message)


class LogFmtFilter(logging.Filter):
    def filter(self, record):
        record.stack_file = ''
        record.err_file = ''

        if record.levelno > 20:  # > INFO
            stack_traceback_fn = registry('StackTracebackFn')
            stack_num = sum(1 for line in open(stack_traceback_fn))
            record.stack_file = f'\n\t File "{stack_traceback_fn}", line {stack_num}, for stacktrace'
            record.err_file = f'\n\t File "{record.pathname}", line {record.lineno}, in {record.funcName}'

        return True


class CustomLog:
    dflt_action = {'debug': False, 'exception': False}
    dflt_severity = {
        'print_traceback': True,
        'warning': copy.deepcopy(dflt_action),
        'error': copy.deepcopy(dflt_action)
    }

    def __init__(self, name, verbosity=logging.INFO):
        self.name = name
        self.verbosity = verbosity

        self.set_default_logger()
        bind(
            registry_log_name.substitute(name=name),
            copy.deepcopy(self.dflt_severity))

        self.logger = logging.getLogger(name)

        self.wrap = LogWrap(self.logger)
        self.logger.warning = self.wrap.warning
        self.logger.error = self.wrap.error

    def get_format(self):
        return logging.Formatter(
            '%(name)s [%(levelname)s]: %(message)s %(err_file)s %(stack_file)s'
        )

    def get_filter(self):
        return LogFmtFilter()

    def get_logger_handler(self):
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(self.verbosity)
        ch.setFormatter(self.get_format())
        ch.addFilter(self.get_filter())
        return ch

    def set_default_logger(self):
        logger = logging.getLogger(self.name)
        logger.setLevel(self.verbosity)
        ch = self.get_logger_handler()
        logger.addHandler(ch)


class LogPlugin(PluginBase):
    @classmethod
    def bind(cls):
        tf = tempfile.NamedTemporaryFile(delete=False)
        bind('StackTracebackFn', tf.name)

        CustomLog('core', logging.WARNING)
        CustomLog('typing', logging.WARNING)
        CustomLog('util', logging.WARNING)
        CustomLog('gear', logging.WARNING)


def core_log():
    return logging.getLogger('core')


def typing_log():
    return logging.getLogger('typing')


def util_log():
    return logging.getLogger('util')


def gear_log():
    return logging.getLogger('gear')



from pygears import registry, bind
from pygears.core.port import OutPort
from pygears.sim import timestep
from pygears.typing_common.codec import code
from pygears.typing import typeof, TLM
from pygears.typing.visitor import TypingVisitorBase
from vcd import VCDWriter
from vcd.gtkw import GTKWSave
from pygears.core.hier_node import HierVisitorBase
from .sim_extend import SimExtend
import os
import fnmatch
import itertools
import atexit


def match(val, include_pattern):
    return any(fnmatch.fnmatch(val, p) for p in include_pattern)


class VCDTypeVisitor(TypingVisitorBase):
    def __init__(self):
        self.fields = {}

    def visit_int(self, type_, field):
        self.fields[field] = type_

    def visit_uint(self, type_, field):
        self.fields[field] = type_

    def visit_unit(self, type_, field):
        self.fields[field] = type_

    def visit_queue(self, type_, field):
        self.visit(type_[0], f'{field}/data')
        self.visit(type_[1:], f'{field}/eot')

    def visit_union(self, type_, field):
        self.visit(type_[0], f'{field}/data')
        self.visit(type_[1], f'{field}/ctrl')

    def visit_default(self, type_, field):
        if hasattr(type_, 'fields'):
            for t, f in zip(type_, type_.fields):
                self.visit(t, f'{field}/{f}')
        else:
            super().visit_default(type_, field)


class VCDValVisitor(TypingVisitorBase):
    def __init__(self, vcd_vars, writer, timestep):
        self.vcd_vars = vcd_vars
        self.writer = writer
        self.timestep = timestep

    def change(self, dtype, field, val):
        self.writer.change(self.vcd_vars[field],
                           timestep() * 10, code(dtype, val))

    def visit_int(self, type_, field, val=None):
        self.change(type_, field, val)

    def visit_union(self, type_, field, val=None):
        self.visit(type_[0], f'{field}/data', val=val[0])
        self.visit(type_[1], f'{field}/ctrl', val=val[1])

    def visit_queue(self, type_, field, val=None):
        val = type_(val)
        self.visit(type_[0], f'{field}/data', val=val[0])
        self.visit(type_[1:], f'{field}/eot', val=val[1:])

    def visit_uint(self, type_, field, val=None):
        self.change(type_, field, val)

    def visit_unit(self, type_, field, val=None):
        self.change(type_, field, val)

    def visit_default(self, type_, field, val=None):
        if hasattr(type_, 'fields'):
            for t, f, v in zip(type_, type_.fields, val):
                self.visit(t, f'{field}/{f}', val=v)


def register_traces_for_intf(dtype, scope, writer):
    vcd_vars = {}

    if typeof(dtype, TLM):
        vcd_vars['data'] = writer.register_var(scope, 'data', 'string')
    else:
        v = VCDTypeVisitor()
        v.visit(dtype, 'data')

        for name, t in v.fields.items():
            vcd_vars[name] = writer.register_var(
                scope, name, 'wire', size=max(int(t), 1))

    for sig in ('valid', 'ready'):
        vcd_vars[sig] = writer.register_var(scope, sig, 'wire', size=1, init=0)

    return vcd_vars


def is_trace_included(port, include, vcd_tlm):
    if not match(f'{port.gear.name}.{port.basename}', include):
        return False

    if (port.dtype is None) or (typeof(port.dtype, TLM) and not vcd_tlm):
        return False

    return True


def module_sav(gtkw, module, vcd_vars):
    gear_vcd_scope = module.name[1:].replace('/', '.')
    with gtkw.group(module.basename):
        for p in itertools.chain(module.out_ports, module.in_ports):
            if isinstance(p, OutPort):
                intf = p.producer
            else:
                intf = p.consumer

            if intf in vcd_vars:
                scope = '.'.join([gear_vcd_scope, p.basename])
                with gtkw.group("    " + p.basename):
                    for name, var in vcd_vars[intf].items():
                        width = ''
                        if var.size > 1:
                            width = f'[{var.size - 1}:0]'

                        gtkw.trace(f'{scope}.{name}{width}')


class VCDHierVisitor(HierVisitorBase):
    def __init__(self, gtkw, writer, include, vcd_tlm):
        self.include = include
        self.vcd_tlm = vcd_tlm
        self.sim_map = registry('SimMap')
        self.gtkw = gtkw
        self.vcd_vars = {}
        self.writer = writer
        self.indent = 0

    def enter_hier(self, name):
        self.gtkw.begin_group(f'{" "*self.indent}{name}', closed=True)
        self.indent += 4

    def exit_hier(self, name):
        self.indent -= 4
        self.gtkw.end_group(f'{" "*self.indent}{name}', closed=True)

    def Gear(self, module):
        self.enter_hier(module.basename)

        if module in self.sim_map:
            gear_vcd_scope = module.name[1:].replace('/', '.')
            for p in itertools.chain(module.out_ports, module.in_ports):
                if not is_trace_included(p, self.include, self.vcd_tlm):
                    continue

                scope = '.'.join([gear_vcd_scope, p.basename])
                if isinstance(p, OutPort):
                    intf = p.producer
                else:
                    intf = p.consumer

                self.vcd_vars[intf] = register_traces_for_intf(
                    p.dtype, scope, self.writer)

                self.enter_hier(p.basename)
                for name, var in self.vcd_vars[intf].items():
                    width = ''
                    if var.size > 1:
                        width = f'[{var.size - 1}:0]'

                    self.gtkw.trace(f'{scope}.{name}{width}')
                self.exit_hier(p.basename)

        super().HierNode(module)

        self.exit_hier(module.basename)

        return True


class VCD(SimExtend):
    def __init__(self, top, fn='pygears.vcd', include=['*'], tlm=False):
        super().__init__()
        self.finished = False
        atexit.register(self.finish)

        outdir = registry('SimArtifactDir')

        vcd_file = open(os.path.join(outdir, fn), 'w')

        self.writer = VCDWriter(vcd_file, timescale='1 ns', date='today')
        bind('VCDWriter', self.writer)
        bind('VCD', self)

        self.clk_var = self.writer.register_var(
            '', 'clk', 'wire', size=1, init=1)

        self.timestep_var = self.writer.register_var(
            '', 'timestep', 'integer', init=0)

        self.handhake = set()

        with open(os.path.join(outdir, 'pygears.gtkw'), 'w') as f:
            gtkw = GTKWSave(f)
            v = VCDHierVisitor(gtkw, self.writer, include, tlm)
            v.visit(top)
            self.vcd_vars = v.vcd_vars

            for intf in self.vcd_vars:
                intf.events['put'].append(self.intf_put)
                intf.events['ack'].append(self.intf_ack)

        sim_map = registry('SimMap')
        for module in sim_map:
            gear_fn = module.name.replace('/', '_')
            with open(os.path.join(outdir, f'{gear_fn}.gtkw'), 'w') as f:
                gtkw = GTKWSave(f)
                module_sav(gtkw, module, self.vcd_vars)

        # sim_map = registry('SimMap')
        # for module, sim_gear in sim_map.items():
        #     gear_vcd_scope = module.name[1:].replace('/', '.')

        #     for p in itertools.chain(module.out_ports, module.in_ports):
        #         if not is_trace_included(p, include, vcd_tlm):
        #             continue

        #         scope = '.'.join([gear_vcd_scope, p.basename])
        #         if isinstance(p, OutPort):
        #             intf = p.producer
        #         else:
        #             intf = p.consumer

        #         self.vcd_vars[intf] = register_traces_for_intf(
        #             p.dtype, scope, self.writer)

        #         intf.events['put'].append(self.intf_put)
        #         intf.events['ack'].append(self.intf_ack)

    def intf_put(self, intf, val):
        v = self.vcd_vars[intf]

        if typeof(intf.dtype, TLM):
            self.writer.change(v['data'], timestep() * 10, str(val))
        else:
            visitor = VCDValVisitor(v, self.writer, timestep() * 10)
            visitor.visit(intf.dtype, 'data', val=val)

        self.writer.change(v['valid'], timestep() * 10, 1)
        return True

    def intf_ack(self, intf):
        v = self.vcd_vars[intf]
        self.writer.change(v['ready'], timestep() * 10, 1)
        self.handhake.add(intf)
        return True

    def before_timestep(self, sim, timestep):
        self.writer.change(self.clk_var, timestep * 10 + 5, 0)
        return True

    def after_timestep(self, sim, timestep):
        self.writer.change(self.timestep_var, timestep * 10, timestep)
        self.writer.change(self.clk_var, timestep * 10, 1)
        for intf, v in self.vcd_vars.items():
            if intf in self.handhake:
                self.writer.change(v['ready'], timestep * 10, 0)
                self.writer.change(v['valid'], timestep * 10, 0)
                self.handhake.remove(intf)

        return True

    def finish(self):
        if not self.finished:
            self.writer.close()
            self.finished = True

    def after_run(self, sim):
        self.finish()


# class SimVCDPlugin(SimPlugin):
#     @classmethod
#     def bind(cls):
#         cls.registry['SimFlow'].append(VCD)



import re

port_list_regex = r'\((?P<port_list>[\s\S]*?)\)'
generic_list_regex = r'(?:#\s*\((?P<param_list>[\s\S]*?)\))'
space_regex = r'\s+'
opt_space_regex = r'\s*'
id_regex_t = r'(?P<{}>[\w"]+)'

module_regex = r'module' + space_regex + id_regex_t.format("name") + \
               opt_space_regex + generic_list_regex + "?" + opt_space_regex + \
               port_list_regex

simple_port_regex = "(?P<dir>input|output)" + opt_space_regex + \
                    id_regex_t.format("type") + \
                    "?" + space_regex + id_regex_t.format("name")

param_regex = id_regex_t.format("name") + \
              space_regex + "=" + space_regex + id_regex_t.format("val")

intf_port_regex = id_regex_t.format("type") + "\." + id_regex_t.format("modport") + \
                  space_regex + id_regex_t.format("name") + opt_space_regex + \
                  "(?:\[(?P<arr_high>[\s\S]*?):(?P<arr_low>[\s\S]*?)\])?"


def extract_ports(port_list):
    port_defs = [p.strip() for p in port_list.split('\n')]

    intfs = []
    ports = []
    for p in port_defs:
        # p = re.sub(r'//.*\n', '\n', p).strip()
        try:
            ret = re.search(simple_port_regex, p)
            if ret:
                ports.append(ret.groupdict())
                if ports[-1]["type"] is None:
                    ports[-1]["type"] = ""
            else:
                ret = re.search(intf_port_regex, p)
                intfs.append(ret.groupdict())
        except:
            pass

    return ports, intfs


def extract_params(param_list):
    if param_list is None:
        return []
    if not param_list.strip():
        return []
    else:
        param_defs = [p.strip() for p in param_list.split(',')]
        params = {}
        for p in param_defs:
            ret = re.search(param_regex, p)
            res = ret.groupdict()
            params[res['name']] = res

        return params


def parse(content):

    content = re.sub(r'//.*\n', '\n', content)
    ret = re.search(module_regex, content)

    ports, intfs = extract_ports(ret["port_list"])
    params = extract_params(ret["param_list"])

    return ret["name"], ports, intfs, params


# with open('/data/projects/pygears/pygears/cookbook/svlib/rng.sv', 'r') as f:
#     print(parse(f.read()))



import collections
import functools


@functools.lru_cache(maxsize=None)
def index_norm_hashable_single(i, size):
    if isinstance(i, tuple):
        start, stop, incr = i

        if start is None:
            start = 0
        elif start < 0:
            start = size + start

        if stop is None:
            stop = size
        elif stop < 0:
            stop += size
        elif stop > size:
            raise IndexError

        if start == stop:
            raise IndexError

        return slice(start, stop, incr)

    else:
        if i < 0:
            i = size + i

        if i > size:
            raise IndexError

        return i


@functools.lru_cache(maxsize=None)
def index_norm_hashable(index, size):
    return tuple(index_norm_hashable_single(i, size) for i in index)


class TemplateArgumentsError(Exception):
    pass


class TemplatedTypeUnspecified(Exception):
    pass


class TypingMeta(type):
    """Base class all types.
    """

    def is_specified(self):
        return True

    def __repr__(self):
        return self.__name__

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return hash(self.__name__)


def type_repr(obj):
    if isinstance(obj, type) and not isinstance(obj, GenericMeta):
        if obj.__module__ == 'builtins':
            return obj.__qualname__
    if obj is ...:
        return ('...')
    return repr(obj)


def type_str(obj):
    if isinstance(obj, type) and not isinstance(obj, GenericMeta):
        if obj.__module__ == 'builtins':
            return obj.__qualname__
    if obj is ...:
        return ('...')
    return str(obj)


class class_and_instance_method:
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls=None):
        if instance is None:
            # return the metaclass method, bound to the class
            type_ = type(cls)
            return getattr(type_, self.func.__name__).__get__(cls, type_)
        return self.func.__get__(instance, cls)


class GenericMeta(TypingMeta):
    """Base class for all types that have a generic parameter.
    """

    def __new__(cls, name, bases, namespace, args=[]):
        if (not bases) or (not hasattr(bases[0],
                                       'args')) or (not bases[0].args):
            # Form a class that has the generic arguments specified
            if isinstance(args, dict):
                namespace.update({
                    '__args__': tuple(args.values()),
                    '__parameters__': tuple(args.keys())
                })
            else:
                namespace.update({'__args__': args})

            spec_cls = super().__new__(cls, name, bases, namespace)
            return spec_cls
        else:
            if len(bases[0].templates) < len(args):
                raise TemplateArgumentsError(
                    "Too many arguments to the templated type: {bases[0]}")

            if isinstance(args, dict):
                for t in args:
                    if t not in bases[0].templates:
                        raise TemplateArgumentsError(
                            f"Template parameter '{t}' not part of the "
                            f"templated type: {bases[0]}")

                tmpl_map = args
            else:
                tmpl_map = {
                    name: val
                    for name, val in zip(bases[0].templates, args)
                }
            return param_subs(bases[0], tmpl_map, {})

    def is_generic(self):
        """Return True if no generic parameter of the type was supplied a value.

        >>> Uint.is_generic()
        True

        >>> Uint[16].is_generic()
        False
        """

        return len(self.args) == 0

    def __bool__(self):
        return self.is_specified()

    def __hash__(self):
        return id(self)

    @functools.lru_cache()
    def is_specified(self):
        """Return True if all generic parameters were supplied concrete values.

        >>> Uint['template'].is_specified()
        False

        >>> Uint[16].is_specified()
        True
        """

        if hasattr(self, '__parameters__'):
            if len(self.args) != len(self.__parameters__):
                return False

        if self.args:
            spec = True
            for a in self.args:
                try:
                    spec &= a.is_specified()
                except AttributeError:
                    if isinstance(a, (str, bytes)):
                        return False
                        # spec &= (templ_var_re.search(a) is None)

            return spec
        else:
            return False

    def __getitem__(self, params):
        if isinstance(params, tuple):
            params = list(params)
        elif not isinstance(params, dict):
            params = [params]

        return self.__class__(
            self.__name__, (self, ) + self.__bases__,
            dict(self.__dict__),
            args=params)

    @property
    def base(self):
        """Returns base generic class of the type.

        >>> assert Uint[16].base == Uint
        """

        if len(self.__bases__) == 1:
            return self
        else:
            return self.__bases__[-2]

    @property
    def templates(self):
        """Returns a list of templated generic variables within the type. The type is
searched recursively. Each template is reported only once.

        >>> Tuple[Tuple['T1', 'T2'], 'T1'].templates
        ['T1', 'T2']
        """

        def make_unique(seq):
            seen = set()
            return [x for x in seq if x not in seen and not seen.add(x)]

        templates = []
        for a in self.args:
            if hasattr(a, 'templates'):
                a_templates = a.templates
                templates += [v for v in a_templates if v not in templates]
            else:
                if isinstance(a, str):  #and templ_var_re.search(a):
                    templates.append(a)

        return make_unique(templates)

    @property
    def args(self):
        """Returns a list of values supplied for each generic parameter.

        >>> Tuple[Uint[1], Uint[2]].args
        [Uint[1], Uint[2]]
        """

        if hasattr(self, '__args__'):
            if hasattr(self, '__default__'):
                plen = len(self.__parameters__)
                alen = len(self.__args__)
                dlen = len(self.__default__)
                missing = plen - alen

                if (missing == 0) or (dlen < missing):
                    return self.__args__
                else:
                    return self.__args__ + self.__default__[-missing:]
            else:
                return self.__args__
        return []

    def __repr__(self):
        if not self.args:
            return super().__repr__()
        else:
            return super().__repr__() + '[%s]' % ', '.join(
                [type_repr(a) for a in self.args])

    def __str__(self):
        if not self.args:
            return super().__repr__()
        else:
            return super().__str__() + '[%s]' % ', '.join(
                [type_str(a) for a in self.args])

    @args.setter
    def args(self, val):
        self.__args__ = val

    @property
    def fields(self):
        """Returns the names of the generic parameters.

        >>> Tuple[Uint[1], Uint[2]].fields
        ('f0', 'f1')

        >>> Tuple[{'u1': Uint[1], 'u2': Uint[2]}].fields
        ('u0', 'u1')
        """

        if hasattr(self, '__parameters__'):
            return self.__parameters__
        else:
            return [f'f{i}' for i in self.keys()]

    def replace(self, field_map, arg_map={}):
        args = {
            field_map.get(k, k): arg_map.get(field_map.get(k, k), v)
            for k, v in zip(self.fields, self.args)
        }

        return self.base[args]

    def __eq__(self, other):
        if not isinstance(other, GenericMeta):
            return False
        elif self.base is not other.base:
            return False
        else:
            if len(self.args) != len(other.args):
                return False
            return all([s == o for s, o in zip(self.args, other.args)])


def param_subs(t, matches, namespace):
    if isinstance(t, bytes):
        t = t.decode()

    # Did we reach the parameter name?
    if isinstance(t, str):
        try:
            return eval(t, namespace, matches)
        except Exception as e:
            return t
            # raise Exception(
            #     f"{str(e)}\n - while evaluating parameter string '{t}'")

    elif isinstance(t, collections.Iterable):
        return type(t)(param_subs(tt, matches, namespace) for tt in t)
    else:
        if isinstance(t, GenericMeta) and (not t.is_specified()):
            args = [
                param_subs(t.args[i], matches, namespace)
                for i in range(len(t.args))
            ]

            if hasattr(t, '__parameters__'):
                args = {name: a for name, a in zip(t.__parameters__, args)}

            return t.__class__(
                t.__name__, t.__bases__, dict(t.__dict__), args=args)

    return t


class EnumerableGenericMeta(GenericMeta):
    """Base class for all types that are iterable.
    """

    def __int__(self):
        """Calculates the bit width of the type.

        >>> int(Tuple[Uint[1], Uint[2]])
        3
        """
        if self.is_specified():
            return sum(map(int, self))
        else:
            raise Exception(
                f"Cannot evaluate width of unspecified generic type"
                f" {type_repr(self)}")

    def __len__(self):
        """The number of elements type generates when iterated.

        >>> Uint[16])
        16
        """
        return len(self.keys())

    def keys(self):
        """Returns a list of keys that can be used for indexing the type.
        """
        return list(range(len(self.args)))

    def index_convert(self, index):
        if isinstance(index, str):
            try:
                return self.fields.index(index)
            except ValueError as e:
                raise KeyError(f'Field "{index}" not in type "{repr(self)}"')
        elif not isinstance(index, slice):
            return index
        else:
            return index.__reduce__()[1]

    def index_norm(self, index):
        if not isinstance(index, tuple):
            return (index_norm_hashable_single(
                self.index_convert(index), len(self)), )
        else:
            return index_norm_hashable(
                tuple(self.index_convert(i) for i in index), len(self))

    def items(self):
        """Generator that yields (key, element) pairs.
        """
        for k in self.keys():
            yield k, self[k]


class Any(metaclass=TypingMeta):
    """Type that can be matched to any other type.
    """
    pass


def typeof(obj, t):
    """Check if a specific type instance is a subclass of the type.

    Args:
       obj: Concrete type instance
       t: Base type class

    """
    try:
        return issubclass(obj, t)
    except TypeError:
        return False



import operator

from pygears import alternative, gear, module
from pygears.typing import Integer, Uint, Int
from pygears.core.intf import IntfOperPlugin
from pygears.util.hof import oper_tree
from pygears.util.utils import gather
from functools import reduce


def mul_type(dtypes):
    length = 0
    for d in dtypes:
        length += int(d)

    if any(issubclass(d, Int) for d in dtypes):
        return Int[length]

    return Uint[length]


@gear(svgen={'svmod_fn': 'mul.sv'}, enablement=b'len(din) == 2')
async def mul(*din: Integer,
              din0_signed=b'typeof(din0, Int)',
              din1_signed=b'typeof(din1, Int)') -> b'mul_type(din)':

    async with gather(*din) as dout:
        yield module().tout(reduce(operator.mul, dout))


@alternative(mul)
@gear
def mul_vararg(*din: Integer, enablement=b'len(din) > 2') -> b'mul_type(din)':
    return oper_tree(din, mul)


class MulIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace']['__mul__'] = mul



from pygears import alternative, TypeMatchError, gear
from pygears.typing import Union
from pygears.common import fmap as common_fmap
from pygears.common import mux, demux


def unionmap_check(dtype, f):
    if not issubclass(dtype, Union):
        return False

    try:
        num_f = len(f)
    except TypeError as e:
        raise TypeMatchError(
            f'Union fmap argument "f" needs to be a sequence, received {f}')

    if len(list(dtype.types)) != num_f:
        raise TypeMatchError(
            'Number of union types different from the number of fmap functions'
        )

    return True


@alternative(common_fmap)
@gear(enablement=b'unionmap_check(din, f)')
def fmap(din, *, f, fdemux=demux(ctrl_out=True), fmux=mux, balance=None):
    demux_dout = din | fdemux
    ctrl = demux_dout[0]
    branches = demux_dout[1:]
    dout = []
    for i, fd in enumerate(f):
        if fd is None:
            if balance is None:
                dout.append(branches[i])
            else:
                dout.append(branches[i] | balance)
        else:
            dout.append(branches[i] | fd)

    if balance is not None:
        ctrl = ctrl | balance

    return fmux(ctrl, *dout)



import asyncio

from pygears import gear, module, GearDone, registry
from pygears.util.find import find
from pygears.sim import delta, clk, sim_log


def decoupler_din_setup(module):
    module.queue = asyncio.Queue(maxsize=module.params['depth'])


@gear(sim_setup=decoupler_din_setup, svgen={'node_cls': None})
async def decoupler_din(din: 'tdin', *, depth) -> None:
    async with din as d:
        await module().queue.put(d)
        while (module().queue.full()):
            await delta()


def decoupler_dout_setup(module):
    module.decoupler_din = find('../decoupler_din')


@gear(sim_setup=decoupler_dout_setup, svgen={'node_cls': None})
async def decoupler_dout(*, t, depth) -> b't':
    queue = module().decoupler_din.queue
    while queue.empty():
        if registry('SimMap')[module().decoupler_din].done:
            raise GearDone

        await clk()

    yield queue.get_nowait()

    queue.task_done()
    await clk()


@gear
def decoupler(din: 'tdin', *, depth=2) -> b'tdin':
    din | decoupler_din(depth=depth)
    return decoupler_dout(t=din.dtype, depth=depth)


buff = decoupler(depth=1)



from pygears import gear, module
from pygears.typing import Queue


@gear(svgen={'svmod_fn': 'project.sv'})
async def project(din: Queue['tdin', 'din_lvl'],
                  *,
                  lvl=1,
                  dout_lvl=b'din_lvl - lvl') -> Queue['tdin', 'dout_lvl']:
    async with din as d:
        yield module().tout(d[:dout_lvl+1])



import operator

from pygears import alternative, gear, module
from pygears.typing import Integer, Int, Uint
from pygears.core.intf import IntfOperPlugin
from pygears.util.hof import oper_tree
from pygears.util.utils import gather
from functools import reduce


def add_type(dtypes):
    max_len = max(int(d) for d in dtypes)
    length = max_len + len(dtypes) - 1

    if (all(issubclass(d, Uint) for d in dtypes)):
        return Uint[length]

    return Int[length]


@gear(svgen={'svmod_fn': 'add.sv'}, enablement=b'len(din) == 2')
async def add(*din: Integer,
              din0_signed=b'typeof(din0, Int)',
              din1_signed=b'typeof(din1, Int)') -> b'add_type(din)':
    async with gather(*din) as dout:
        yield module().tout(reduce(operator.add, dout))


@alternative(add)
@gear
def add_vararg(*din, enablement=b'len(din) > 2') -> b'add_type(din)':
    return oper_tree(din, add)


class AddIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace']['__add__'] = add



from pygears.svgen.svmod import SVModuleGen
from pygears.typing import Uint, Tuple
from pygears.svgen.inst import SVGenInstPlugin
from pygears.common.quenvelope import quenvelope


class SVGenQuEnvelope(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_sv_port_config(self, modport, type_, name):
        cfg = super().get_sv_port_config(modport, type_, name)

        fields = {
            'data': type_[0],
            'subenvelope': Uint[type_.lvl - self.lvl],
            'out_eot': Uint[self.lvl]
        }

        cfg['local_type'] = Tuple[fields]
        cfg['lvl'] = type_.lvl

        return cfg

    def get_module(self, template_env):
        self.lvl = self.node.params['lvl']
        intfs = list(self.sv_port_configs())

        if intfs[0]['lvl'] > self.lvl:
            dout_align = f'&din_s.subenvelope'
        else:
            dout_align = 1

        context = {
            'comment': self.__doc__,
            'dout_align': dout_align,
            'lvl': self.lvl,
            'module_name': self.sv_module_name,
            'intfs': intfs
        }
        return template_env.render_local(__file__, "quenvelope.j2", context)


class SVGenQuEnvelopePlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][quenvelope] = SVGenQuEnvelope



import logging
import math
import os
import socket
from subprocess import DEVNULL, Popen

import jinja2

from pygears import registry
from pygears.definitions import ROOT_DIR
from pygears.sim import sim_log
from pygears.sim.extens.rand_base import RandBase
from pygears.sim.modules.sim_socket import SimSocket, u32_bytes_decode
from pygears.svgen.util import svgen_typedef
from pygears.util.fileio import save_file


class SVRandConstraints:
    def __init__(self,
                 name='dflt',
                 cons=[],
                 dtype=None,
                 cls='dflt_tcon',
                 cls_params=None):
        self.name = name
        self.cons = cons.copy()
        self.dtype = dtype
        self.cls = cls
        self.cls_params = cls_params
        self.cvars = {}
        self.cvars[name] = svgen_typedef(dtype, name)

    def add_var(self, name, dtype):
        self.cvars[name] = svgen_typedef(dtype, name)


class SVRandSocket(RandBase):
    SVRAND_CONN_NAME = "_svrand"

    def __init__(self, top, cons, run=False, port=4567, **kwds):
        super().__init__(top, cons, **kwds)
        self.run_cosim = run
        self.cosim_pid = None
        kwds['batch'] = True
        kwds['clean'] = True
        self.kwds = kwds
        self.port = port
        self.open_sock = True

    def create_type_cons(self, desc={}):
        tcons = SVRandConstraints(
            name=desc['name'],
            dtype=desc['dtype'],
            cons=desc['cons'],
            cls=desc['cls'],
            cls_params=desc['cls_params'])

        for name, dtype in desc['params'].items():
            tcons.add_var(name, dtype)

        return tcons

    def before_setup(self, sim):
        sim_map = registry('SimMap')
        for module, sim_gear in sim_map.items():
            if isinstance(sim_gear, SimSocket):
                self.open_sock = False

        self.create_svrand_top()

        if self.open_sock:
            self.cosim_pid = None
            if self.run_cosim:
                outdir = registry('SimArtifactDir')
                args = ' '.join(f'-{k} {v if not isinstance(v, bool) else ""}'
                                for k, v in self.kwds.items()
                                if not isinstance(v, bool) or v)
                if 'seed' in self.kwds:
                    sim_log().warning(
                        'Separately set seed for cosimulator. Ignoring SimRandSeed.'
                    )
                else:
                    args += f' -seed {registry("SimRandSeed")}'

                if sim_log().isEnabledFor(logging.DEBUG):
                    stdout = None
                else:
                    stdout = DEVNULL

                sim_log().info(f'Running cosimulator with: {args}')
                self.cosim_pid = Popen(
                    [f'./svrand_runsim.sh'] + args.split(' '),
                    stdout=stdout,
                    stderr=stdout,
                    cwd=outdir)
                sim_log().info(f"Cosimulator started")

            self.connect()
        else:
            hooks = {}
            hooks['module_init'] = 'svrand_top rand_i();'
            data_rng = [x for x in range(1, len(self.constraints) + 1)]
            hooks[
                'synchro_req'] = f'if (data inside {{{", ".join(str(x) for x in data_rng)}}}) ret = rand_i.get_rand(synchro_handle, data);'
            registry('SimConfig')['SimSocketHooks'] = hooks

    def connect(self):
        # Create a TCP/IP socket
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # Bind the socket to the port
        server_address = ('localhost', self.port)
        print('starting up on %s port %s' % server_address)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind(server_address)

        if self.cosim_pid:
            self.sock.settimeout(1)
        else:
            self.sock.setblocking(True)

        self.sock.listen(1)

        if self.cosim_pid:
            ret = None
            while ret is None:
                try:
                    conn, addr = self.sock.accept()
                    break
                except socket.timeout:
                    ret = self.cosim_pid.poll()
                    if ret is not None:
                        sim_log().error(f"Cosimulator error: {ret}")
                        raise Exception

        else:
            sim_log().info("Wait for connection")
            self.sock.listen(1)
            conn, addr = self.sock.accept()

        msg = conn.recv(1024)
        port_name = msg.decode()

        self.conn = conn

        print(f"Connection received for {port_name}")

    def parse_name(self, name):
        for i, c in enumerate(self.constraints):
            if c.name == name:
                dtype = c.dtype
                req = i + 1
                break
        return req, dtype

    def send_req(self, req, dtype):
        data = None

        # Send request
        pkt = req.to_bytes(4, byteorder='little')
        self.conn.sendall(b'\x01\x00\x00\x00' + pkt)

        # Get random data
        while data is None:
            try:
                buff_size = math.ceil(int(dtype) / 8)
                if buff_size < 4:
                    buff_size = 4
                if buff_size % 4:
                    buff_size += 4 - (buff_size % 4)
                data = self.conn.recv(buff_size)
            except socket.error:
                sim_log().error(f'socket error on {self.SYNCHRO_HANDLE_NAME}')
                raise socket.error
        data = u32_bytes_decode(data, dtype)
        return data

    def get_rand(self, name):
        req, dtype = self.parse_name(name)
        data = self.send_req(req, dtype)
        return data

    def at_exit(self, sim):
        if self.open_sock:
            if hasattr(self, 'conn'):
                self.conn.sendall(b'\x00\x00\x00\x00')
                self.conn.close()
                self.sock.close()
                self.open_sock = False

            if self.cosim_pid is not None:
                self.cosim_pid.terminate()

    def create_svrand_top(self):
        base_addr = os.path.dirname(__file__)
        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(base_addr),
            trim_blocks=True,
            lstrip_blocks=True)
        env.globals.update(
            zip=zip, int=int, print=print, issubclass=issubclass)

        context = {
            'tcons': self.constraints,
            'open_sock': self.open_sock,
            'port': self.port
        }

        res = env.get_template('svrand_top.j2').render(context)
        save_file('svrand_top.sv', self.outdir, res)

        # custom classes
        for con in self.constraints:
            if con.cls == 'qenvelope':
                context = {'tcon': con}
                res = env.get_template('qenvelope.j2').render(context)
                save_file(f'qenvelope_{con.name}.sv', self.outdir, res)

        context = {
            'dti_verif_path':
            os.path.abspath(os.path.join(ROOT_DIR, 'sim', 'dpi')),
            'out_path':
            os.path.abspath(self.outdir),
            'includes': [os.path.abspath(os.path.join(self.outdir, '*.sv'))]
        }

        res = env.get_template('runsim.j2').render(context)
        fname = save_file('svrand_runsim.sh', self.outdir, res)
        os.chmod(fname, 0o777)



from pygears.typing.visitor import TypingVisitorBase
from pygears.typing import Tuple, Uint


def dtype_mask(dtype):
    return ((1 << int(dtype)) - 1)


class TypeCode(TypingVisitorBase):
    def visit_uint(self, dtype, field=None, data=None):
        return data & dtype_mask(dtype)

    def visit_bool(self, dtype, field=None, data=None):
        return data

    def visit_int(self, dtype, field=None, data=None):
        return data & dtype_mask(dtype)

    def visit_queue(self, dtype, field=None, data=None):
        return self.visit_tuple(
            Tuple[(dtype[0], ) + (Uint[1], ) * dtype.lvl], data=data)

    def visit_tuple(self, dtype, field=None, data=None):
        ret = 0
        if len(dtype) != len(data):
            raise ValueError

        for d, t in zip(reversed(data), reversed(dtype)):
            field_data = self.visit(t, data=d)
            if field_data is not None:
                ret <<= int(t)
                ret |= field_data

        return ret

    def visit_array(self, dtype, field=None, data=None):
        return self.visit_tuple(Tuple[(dtype[0], ) * len(dtype)], data=data)

    def visit_union(self, dtype, field=None, data=None):
        return self.visit_tuple(Tuple[(dtype[0], dtype[1])], data=data)


class TypeDecode(TypingVisitorBase):
    def visit_int(self, dtype, field=None, data=None):
        if data.bit_length() == int(dtype):
            return dtype(data - (1 << int(dtype)))
        else:
            return dtype(data)

    def visit_bool(self, dtype, field=None, data=None):
        return dtype(data)

    def visit_uint(self, dtype, field=None, data=None):
        return dtype(data & dtype_mask(dtype))

    def visit_tuple(self, dtype, field=None, data=None):
        ret = []
        for t in dtype:
            ret.append(self.visit(t, data=data & dtype_mask(t)))
            data >>= int(t)

        return dtype(tuple(ret))

    def visit_unit(self, dtype, field=None, data=None):
        return None

    def visit_queue(self, dtype, field=None, data=None):
        sub_data_mask = ((1 << (int(dtype) - 1)) - 1)
        sub_data = data & sub_data_mask

        ret = self.visit(dtype[:-1], data=sub_data)

        eot = bool(data & (1 << (int(dtype) - 1)))

        if dtype.lvl == 1:
            return dtype((ret, eot))
        else:
            return dtype((ret[0], *ret[1:], eot))

    def visit_array(self, dtype, field=None, data=None):
        ret = []
        for t in dtype:
            ret.append(self.visit(t, data=data & dtype_mask(t)))
            data >>= int(t)

        return dtype(tuple(ret))

    def visit_union(self, dtype, field=None, data=None):
        ret = []
        for t in dtype:
            ret.append(self.visit(t, data=data & dtype_mask(t)))
            data >>= int(t)

        return dtype(tuple(ret))
        # return self.visit_tuple(Tuple[(dtype[0], dtype[1])], data=data)


def code(dtype, data):
    return TypeCode().visit(dtype, data=data)


def decode(dtype, data):
    return TypeDecode().visit(dtype, data=data)



from itertools import islice

from pygears import registry
from pygears.core.util import perpetum
from pygears.sim import sim_log
from pygears.sim.extens.scvrand import SCVRand
from pygears.sim.extens.svrand import SVRandSocket
from pygears.typing import Queue
from pygears.typing.queue import QueueMeta


def get_rand(name, cnt=None):
    randomizator = registry('SimConfig')['Randomizator']

    if isinstance(randomizator, SVRandSocket):
        if randomizator.open_sock:
            rand_func = perpetum(randomizator.get_rand, name)
        else:
            req, dtype = randomizator.parse_name(name)
            simsoc = registry('SimConfig')['SimSocket']
            rand_func = perpetum(simsoc.send_req, req, dtype)
    elif isinstance(randomizator, SCVRand):
        rand_func = perpetum(randomizator.get_rand, name)
    else:
        sim_log().error('Randomizator not set')
        return None

    if cnt is not None:
        yield from islice(rand_func, cnt)
    else:
        yield from rand_func


def rand_seq(name, cnt=None):
    randomizator = registry('SimConfig')['Randomizator']
    dtype = randomizator.get_dtype_by_name(name)

    if isinstance(dtype, Queue) or isinstance(dtype, QueueMeta):
        rnd_data = get_rand(f'{name}_data')
        rnd_eot = get_rand(f'{name}_eot')
        tout = None
        while cnt != 0:
            eot = next(rnd_eot)
            data = next(rnd_data)
            if tout is None:
                tout = Queue[type(data), len(eot)]

            yield tout((data, *eot))
            if cnt is not None:
                if eot == int('1' * len(eot), 2):
                    cnt -= 1
    else:
        yield from get_rand(name, cnt)


class ConstraintWrap:
    def __init__(self,
                 dtype,
                 name,
                 cons=[],
                 params={},
                 cls='dflt_tcon',
                 cls_params=None,
                 eot_cons=[],
                 eot_cls='qenvelope',
                 eot_cls_params=None,
                 eot_params={}):
        self.dtype = dtype
        self.name = name

        self.cons = cons
        self.params = params
        self.cls = cls
        self.cls_params = cls_params

        self.is_queue = False
        if isinstance(dtype, Queue) or isinstance(dtype, QueueMeta):
            self.is_queue = True
            self.eot_cons = eot_cons
            self.eot_params = eot_params
            self.eot_cls = eot_cls
            self.eot_cls_params = eot_cls_params

    def get_data_desc(self):
        return {
            'name': self.name if not self.is_queue else f'{self.name}_data',
            'dtype': self.dtype if not self.is_queue else self.dtype[0],
            'cons': self.cons,
            'params': self.params,
            'cls': self.cls,
            'cls_params': self.cls_params
        }

    def get_eot_desc(self):
        return {
            'name': f'{self.name}_eot',
            'dtype': self.dtype.eot,
            'cons': self.eot_cons,
            'params': self.eot_params,
            'cls': self.eot_cls,
            'cls_params': self.eot_cls_params
        }


def create_constraint(dtype,
                      name,
                      cons=[],
                      params={},
                      cls='dflt_tcon',
                      cls_params=None,
                      eot_cons=[],
                      eot_cls='qenvelope',
                      eot_cls_params=None,
                      eot_params={}):
    return ConstraintWrap(**locals())



from pygears import registry
from pygears.sim import sim_log
from pygears.sim.extens.sim_extend import SimExtend
from pygears.typing import Queue


class RandBase(SimExtend):
    def __init__(self, top, cons, **kwds):
        super().__init__()
        self.outdir = registry('SimArtifactDir')
        self.constraints = self.create_constraints(cons)
        registry('SimConfig')['Randomizator'] = self

    def get_dtype_by_name(self, name):
        for constraint in self.constraints:
            if constraint.name == name:
                return constraint.dtype

        # check if queue
        data = None
        eot = None
        for constraint in self.constraints:
            if constraint.name == f'{name}_data':
                data = constraint.dtype
            if constraint.name == f'{name}_eot':
                eot = constraint.dtype

        if data and eot:
            return Queue[data, len(eot)]

    def create_constraints(self, cons):
        constraints = []
        for c in cons:
            constraints.append(self.create_type_cons(c.get_data_desc()))
            if c.is_queue:
                constraints.append(self.create_type_cons(c.get_eot_desc()))
        return constraints

    def create_type_cons(self, desc={}):
        sim_log().error('Create type constraints function not implemented.')

    def get_rand(self, name):
        sim_log().error('Get rand function not implemented.')



from pygears.typing import Queue
from pygears.svgen.generate import isinput


def din_data_cat(intfs):
    data = []
    for intf in isinput(intfs):
        if issubclass(intf['type'], Queue):
            if int(intf['type'][0]) > 0:
                data.append(f'{intf["name"]}_s.data')
        else:
            if int(intf['type']) > 0:
                data.append(f'{intf["name"]}_s')

    return f'{{ {", ".join(reversed(data))} }}'



import functools
import inspect
import sys
import traceback
from .err import enum_traceback


class MultiAlternativeError(Exception):
    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        ret = ['\n']
        for func, e, info in self.errors:
            uwrp = inspect.unwrap(
                func, stop=(lambda f: hasattr(f, "__signature__")))
            fn = inspect.getfile(uwrp)
            _, ln = inspect.getsourcelines(uwrp)
            ret.extend(enum_traceback(info[2]))
            # ret.extend(traceback.format_tb(info[2]))
            ret.append(f'  File "{fn}", line {ln}, in {uwrp.__name__}\n')
            ret.extend(traceback.format_exception_only(*info[:2]))
            ret.append('\n')
        return str(''.join(ret))


def argspec_unwrap(func):
    # uwrp = inspect.unwrap(func, stop=(lambda f: hasattr(f, "__signature__")))
    # return inspect.getfullargspec(uwrp)
    return inspect.getfullargspec(func)


def extract_arg_kwds(kwds, func):
    arg_names, _, _, _, kwonlyargs, *_ = argspec_unwrap(func)

    arg_kwds = {}
    kwds_only = {}
    for k in list(kwds.keys()):
        if k in arg_names:
            arg_kwds[k] = kwds[k]
        elif k in kwonlyargs:
            kwds_only[k] = kwds[k]
        else:
            raise TypeError(
                f"{func.__name__}() got an unexpected keyword argument '{k}'")

    return arg_kwds, kwds_only


def combine_arg_kwds(args, kwds, func):
    arg_names, varargs, *_ = argspec_unwrap(func)

    if varargs:
        return args

    args = list(args)
    args_comb = []
    for a in arg_names:
        if a in kwds:
            args_comb.append(kwds[a])
        elif args:
            args_comb.append(args.pop(0))
        else:
            break

    # If some args could not be matched to argument names, raise an Exception
    if args:
        raise TypeError(f"Too many positional arguments for {func.__name__}()")

    return args_comb


def all_args_specified(args, func):
    arg_names, varargs, _, _, kwds, _, types = argspec_unwrap(func)
    if varargs:
        return len(args) > 0
    elif len(args) == len(arg_names):
        return True
    else:
        return False


class Partial:
    '''The Partial class implements a mechanism similar to that of the
    functools.partial, with one important difference with how calling
    its objects operates.

    When Partial class is instantiated, arguments that are to be passed to the
    wrapped function are inspected. If not enough positional arguments, as
    demanded by the wrapped function, have been supplied, a Partial object is
    returned that remembers which arguments have been supplied. Otherwise, if
    enough arguments have been supplied, instead of returning a new Partial
    object, the wrapped function is called with the supplied arguments and its
    return value is returned instead.

    When instantiated Partial object is called (via __call__ method), the
    arguments supplied to the call are combined with the arguments already
    supplied via Partial object constructor and Partial class is then called
    with a combined set of arguments to either return a new Partial object or
    call the wrapped function as described by the above paragraph

    This class also supports supplying arguments by pipe '|' operator.

    '''

    def __init__(self, func, *args, **kwds):
        functools.update_wrapper(self, func)
        self.func = func
        self.args = args
        self.kwds = kwds

    def __call__(self, *args, **kwds):
        prev_kwds = self.kwds.copy()
        prev_kwds.update(kwds)
        kwds = prev_kwds

        args = self.args + args

        alternatives = [self.func] + getattr(self.func, 'alternatives', [])
        errors = []

        for func in alternatives:
            try:
                kwd_intfs, kwd_params = extract_arg_kwds(kwds, func)
                args_comb = combine_arg_kwds(args, kwd_intfs, func)

                if all_args_specified(args_comb, func):
                    argspec = argspec_unwrap(func)
                    if '__base__' in argspec.kwonlyargs:
                        kwd_params['__base__'] = self.func
                    return func(*args_comb, **kwd_params)
            except Exception as e:
                # If no alternatives, just re-raise an error
                if len(alternatives) == 1:
                    raise e
                else:
                    errors.append((func, e, sys.exc_info()))
        else:
            if len(errors) == len(alternatives):
                raise MultiAlternativeError(errors)
            else:
                # If some alternative can handle more arguments, try to wait
                # for it
                return Partial(self.func, *args, **kwds)

    def __matmul__(self, iin):
        return self(intfs=iin)

    def __or__(self, iin):
        return iin.__ror__(self)

    def __ror__(self, iin):
        if type(iin) == tuple:
            return self(*iin)
        else:
            return self(iin)


class Definition:
    '''The Definition class postpones creation of the Partial object on a function
until the function object (definition) has been called to assign first set (or
all) of arguments. This allows for creation of separate Partial objects for
each of the function invocations.

    This class also supports supplying arguments by pipe '|' operator.

    '''

    def __init__(self, func):
        self.func = func

        functools.update_wrapper(self, func)

    def __call__(self, *args, **kwds):
        return Partial(self.func, *args, **kwds)

    def __or__(self, iin):
        module = Partial(self.func)
        return module | iin

    def __ror__(self, iin):
        if isinstance(iin, tuple):
            return Partial(self.func, *iin)
        else:
            return Partial(self.func, iin)



from pygears import gear
from pygears.typing import Uint, Queue, Array


TDin = Array[Uint['w_data'],'no']
TOut = Queue[Uint['w_data']]

@gear(outnames=['dout'])
def width_reductor(din: TDin) -> TOut:
    pass



from pygears.svgen.inst import SVGenInstPlugin
from pygears.svgen.svmod import SVModuleGen
from pygears.common.union import union_sync
from .syncguard import SVGenSyncGuard


class SVGenUnionSyncBase(SVModuleGen):
    def __init__(self, node):
        super().__init__(node)

        if 'outsync' not in self.node.params:
            self.node.params['outsync'] = True

        if self.node.params['outsync']:
            self.syncguard = SVGenSyncGuard(f'{self.sv_module_name}_syncguard',
                                            len(self.node.in_ports))
        else:
            self.syncguard = None

    @property
    def is_generated(self):
        return True

    @property
    def sv_file_name(self):
        if self.syncguard is None:
            return super().sv_file_name
        else:
            return super().sv_file_name, self.syncguard.sv_file_name

    def get_module(self, template_env, template_fn):

        context = {
            'outsync': self.node.params['outsync'],
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs()),
            'ctrl': self.node.params['ctrl']
        }
        contents = template_env.render_local(__file__, template_fn, context)

        if self.syncguard is None:
            return contents
        else:
            return contents, self.syncguard.get_module(template_env)


class SVGenUnionSync(SVGenUnionSyncBase):
    def get_module(self, template_env):

        return super().get_module(template_env, 'union_sync.j2')


class SVGenUnionPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][union_sync] = SVGenUnionSync



import os

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
COMMON_SVLIB_DIR = os.path.join(ROOT_DIR, 'common', 'svlib')
COOKBOOK_SVLIB_DIR = os.path.join(ROOT_DIR, 'cookbook', 'svlib')
USER_SVLIB_DIR = os.path.expanduser('~/.pygears/svlib')



import pkg_resources

__version__ = pkg_resources.get_distribution("pygears").version

import sys
from asyncio import CancelledError as GearDone
from asyncio.queues import QueueEmpty

from pygears.registry import PluginBase, bind, registry, clear
from pygears.util.find import find

import pygears.sim

from pygears.core.err import pygears_excepthook, ErrReportLevel
from pygears.core.type_match import TypeMatchError
from pygears.registry import PluginBase, bind, registry, clear
from pygears.core.gear import gear, alternative, module
from pygears.core.intf import Intf
from pygears.core.partial import MultiAlternativeError

import pygears.common
import pygears.typing
import pygears.typing_common

from pygears.core.log import LogPlugin
# import os
# from pygears.registry import load_plugin_folder
# load_plugin_folder(os.path.join(os.path.dirname(__file__), 'common'))

sys.excepthook = pygears_excepthook

__all__ = [
    'registry', 'ErrReportLevel', 'bind', 'gear', 'clear', 'Intf',
    'PluginBase', 'find', 'MultiAlternativeError', 'GearDone', 'QueueEmpty',
    'module'
]



from pygears.typing.uint import IntMeta, UintMeta
from pygears.typing import Queue
from pygears.svgen.svgen import SVGenPlugin
from pygears.svgen.util import svgen_visitor
from pygears.rtl.connect import rtl_connect
from pygears.rtl.inst import RTLNodeInstPlugin
from pygears.svgen.svmod import SVModuleGen
from pygears.common.cast import cast
from pygears.rtl.gear import RTLGearHierVisitor


class SVGenCast(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        outtype = self.node.out_ports[0].dtype
        intype = self.node.in_ports[0].dtype
        out_width = int(outtype)

        if isinstance(outtype, IntMeta):
            if isinstance(intype, UintMeta):
                data_expr = f"{out_width}'(din.data)"
            else:
                data_expr = f"{out_width}'(signed'(din.data))"
        elif issubclass(outtype, Queue) and outtype[0] == intype:
            if (int(intype) != 0):
                data_expr = "{1'b1, din.data}"
            else:
                data_expr = "1"
        else:
            data_expr = f"{out_width}'(din.data)"

        context = {
            'data_expr': data_expr,
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs())
        }

        return template_env.render_local(__file__, "cast.j2", context)


@svgen_visitor
class RemoveEqualReprCastVisitor(RTLGearHierVisitor):
    def cast(self, node):
        pout = node.out_ports[0]
        pin = node.in_ports[0]

        if int(pin.dtype) == int(pout.dtype):
            node.bypass()


class SVGenSievePlugin(RTLNodeInstPlugin, SVGenPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][cast] = SVGenCast
        cls.registry['SVGenFlow'].insert(
            cls.registry['SVGenFlow'].index(rtl_connect) + 1,
            RemoveEqualReprCastVisitor)



from pygears.sim.sim_gear import SimGear
from pygears.sim import clk, timestep, delta
from pygears import GearDone, module
import time


class CosimNoData(Exception):
    pass


class CosimBase(SimGear):
    SYNCHRO_HANDLE_NAME = "_synchro"

    def __init__(self, gear, timeout=-1):
        super().__init__(gear)
        self.timeout = timeout

    def finish(self):
        super().finish()

    async def func(self, *args, **kwds):
        activity_monitor = 0
        din_pulled = set()
        dout_put = set()
        try:
            while True:
                activity_monitor += 1
                if activity_monitor == self.timeout:
                    raise GearDone

                self.handlers[self.SYNCHRO_HANDLE_NAME].cycle()

                for p in self.gear.in_ports:
                    intf = p.consumer
                    if p.basename not in self.handlers:
                        continue

                    hin = self.handlers[p.basename]

                    if intf.done():
                        hin.close()
                        del self.handlers[p.basename]

                    if p not in din_pulled:
                        if not intf.empty():
                            data = intf.pull_nb()
                            din_pulled.add(p)
                            hin.send(data)
                        else:
                            hin.reset()

                self.handlers[self.SYNCHRO_HANDLE_NAME].forward()

                for p in self.gear.out_ports:
                    intf = p.producer
                    if intf.ready_nb():
                        hout = self.handlers[p.basename]
                        try:
                            intf.put_nb(hout.read())
                            dout_put.add(p)
                            activity_monitor = 0
                        except CosimNoData:
                            pass
                    else:
                        hout.reset()

                await delta()

                for p in dout_put.copy():
                    intf = p.producer
                    if intf.ready_nb():
                        hout = self.handlers[p.basename]
                        hout.ack()
                        dout_put.remove(p)

                self.handlers[self.SYNCHRO_HANDLE_NAME].back()

                for p in din_pulled.copy():
                    if p.basename not in self.handlers:
                        continue

                    hin = self.handlers[p.basename]
                    if hin.ready():
                        activity_monitor = 0
                        din_pulled.remove(p)
                        intf = p.consumer
                        intf.ack()

                await clk()

        except GearDone as e:
            # print(f"SimGear canceling: {self.gear.name}")
            for p in self.gear.out_ports:
                p.producer.finish()

            self.finish()
            raise e



import logging
import sys


def init_loger(fn):
    logging.basicConfig(
        filename=fn,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        level=logging.INFO,
        datefmt='%d-%m-%Y %H:%M:%S')
    logger = logging.getLogger()

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.INFO)
    logger.addHandler(ch)



from pygears.svgen.svmod import SVModuleGen
from pygears.typing.queue import Queue
from pygears.svgen.inst import SVGenInstPlugin
from pygears.common.mux import mux_valve


class SVGenMuxValve(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):

        intf_cfgs = list(self.sv_port_configs())
        context = {
            'module_name': self.sv_module_name,
            'intfs': intf_cfgs
        }
        return template_env.render_local(__file__, "mux_valve.j2", context)


class SVGenMuxValvePlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][mux_valve] = SVGenMuxValve



from pygears import alternative, gear
from pygears.typing import Queue, Uint


@gear
def take(din: Queue['T'], cfg: Uint['N']) -> b'Queue[T]':
    pass


@alternative(take)
@gear(svgen={'svmod_fn': 'qtake.sv'})
def qtake(din: Queue['Tdin', 2], cfg: Uint['N']) -> Queue['Tdin', 2]:
    '''
    Takes given number of queues. Number given by cfg.
    Counts lower eot. Higher eot resets.
    '''
    pass



class Port:
    def __init__(self, gear, index, basename, producer=None, consumer=None):
        self.gear = gear
        self.index = index
        self.producer = producer
        self.consumer = consumer
        self.basename = basename

    @property
    def dtype(self):
        if self.producer is not None:
            return self.producer.dtype
        else:
            return self.consumer.dtype

    def get_queue(self, port=None):
        if port is None:
            port = self

        return self.producer.get_consumer_queue(port)

    def finish(self):
        self.consumer.finish()


class InPort(Port):
    pass


class OutPort(Port):
    pass
